import time

import requests
import requests.packages.urllib3.exceptions

from .logger import setup_logger

# Disable the InsecureRequestWarning
requests.packages.urllib3.disable_warnings(
    category=requests.packages.urllib3.exceptions.InsecureRequestWarning
)

logger = setup_logger(name=__name__)


def get_user_agent() -> str:
    return "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/46.0.2490.86 Safari/537.36"


def request_(
    method: str,
    url: str,
    retries: int = 10,
    wait_time: float = 5,
    silence_logs: bool = False,
    **kwargs,
) -> requests.Response:
    """
    Submit to requests module with retry and error logic.

    :param method: (str) ['get', 'put', 'post', 'delete', 'patch', 'head', 'options']
    :param url: (str) URL to call.
    :param retries: (int) If an error is reached how many times to re-attempt request.
    :param wait_time: (float) Time to wait between re-attempts, when necessary.
    :param silence_logs: (bool) Suppress error messages in logs.  (Default: False)
    :param kwargs: Options:
        params: (optional) Dictionary, list of tuples or bytes to send in the query string for the :class:`Request`.
        data: (optional) Dictionary, list of tuples, bytes, or file-like object to send in the body of the
            :class:`Request`.
        json: (optional) A JSON serializable Python object to send in the body of the :class:`Request`.
        headers: (optional) Dictionary of HTTP Headers to send with the :class:`Request`.
        cookies: (optional) Dict or CookieJar object to send with the :class:`Request`.
        files: (optional) Dictionary of ``'name': file-like-objects`` (or ``{'name': file-tuple}``) for multipart
            encoding upload. ``file-tuple`` can be a 2-tuple ``('filename', fileobj)``, 3-tuple ``('filename',
            fileobj, 'content_type')`` or a 4-tuple ``('filename', fileobj, 'content_type', custom_headers)``,
            where ``'content-type'`` is a string defining the content type of the given file and ``custom_headers``
            a dict-like object containing additional headers to add for the file.
        auth: (optional) Auth tuple to enable Basic/Digest/Custom HTTP Auth.
        timeout: (optional) How many seconds to wait for the server to send data before giving up, as a float,
            or a :ref:`(connect timeout, read timeout) <timeouts>` tuple. :type timeout: float or tuple
        allow_redirects: (optional) Boolean. Enable/disable GET/OPTIONS/POST/PUT/PATCH/DELETE/HEAD redirection.
            Defaults to ``True`` :type allow_redirects: bool
        proxies: (optional) Dictionary mapping protocol to the URL of the proxy.
        verify: (optional) Either a boolean, in which case it controls whether we verify the server's TLS
            certificate, or a string, in which case it must be a path to a CA bundle to use. Defaults to ``True``.
        stream: (optional) if ``False``, the response content will be immediately downloaded.
        cert: (optional) if String, path to ssl client cert file (.pem). If Tuple, ('cert', 'key') pair.

    :return: :class:`Response <Response>` object :rtype: requests.Response
    """
    retry_attempts = 0
    result = None
    while retry_attempts <= retries:
        try:
            passable_options = [
                "params",
                "data",
                "json",
                "headers",
                "cookies",
                "files",
                "auth",
                "allow_redirects",
                "proxies",
                "verify",
                "stream",
                "cert",
                "timeout",
            ]
            result = requests.request(
                method=method.upper(),
                url=url,
                **{k: v for k, v in kwargs.items() if k in passable_options},
            )
            if result.status_code < 400:
                break  # Only stop looping if the status_code is reported as not an error.
        except requests.exceptions.SSLError:
            if not silence_logs:
                logger.debug("Disabling SSL verification for the next request attempt.")
            kwargs.update(
                {
                    "verify": False,
                }
            )
            continue  # Skip the wait and try again but with SSL verification off.
        except requests.exceptions.RequestException as e:
            if not silence_logs:
                logger.error(f"Encountered error: {e}")
        except requests.packages.urllib3.exceptions as e:
            if not silence_logs:
                logger.error(f"Encountered error: {e}")
        if not silence_logs:
            logger.info(
                f"Retrying request in {wait_time}s.  Retries remaining: {retries - retry_attempts}"
            )
        retry_attempts += 1
        time.sleep(wait_time)

    if result.status_code == 400:
        if not silence_logs:
            logger.info(
                f"Status code 400 indicates that the server cannot or will not process "
                f"the request due to something that is perceived to be a client error."
            )
    elif result.status_code == 401:
        if not silence_logs:
            logger.info(
                f"Status code 401 indicates the client request has not been completed "
                f"because it lacks valid authentication credentials for the requested resource."
            )
    elif result.status_code == 404:
        if not silence_logs:
            logger.info(
                f"Status code 404 indicates that the server cannot find the requested resource."
            )
    elif result.status_code == 405:
        if not silence_logs:
            logger.info(
                f"Status code 405 indicates that the server knows the request method, "
                f"but the target resource doesn't support this method."
            )
    elif result.status_code == 415:
        if not silence_logs:
            logger.info(
                f"Status code 415 indicates that the server refuses to accept the request "
                f"because the payload format is in an unsupported format; so stop trying!"
            )
    elif result.status_code == 429:
        if not silence_logs:
            logger.info(
                f"Status code 429 indicates the user has sent too many requests in a "
                f"given amount of time."
            )

    return result



valid_category_ids = [
    "ANY",
    "NONE",
    "OTHER_ADULT_MATERIAL",
    "ADULT_THEMES",
    "LINGERIE_BIKINI",
    "NUDITY",
    "PORNOGRAPHY",
    "SEXUALITY",
    "ADULT_SEX_EDUCATION",
    "K_12_SEX_EDUCATION",
    "SOCIAL_ADULT",
    "OTHER_BUSINESS_AND_ECONOMY",
    "CORPORATE_MARKETING",
    "FINANCE",
    "PROFESSIONAL_SERVICES",
    "CLASSIFIEDS",
    "TRADING_BROKARAGE_INSURANCE",
    "CUSTOM_00",
    "CUSTOM_01",
    "CUSTOM_02",
    "CUSTOM_03",
    "CUSTOM_04",
    "CUSTOM_05",
    "CUSTOM_06",
    "CUSTOM_07",
    "CUSTOM_08",
    "CUSTOM_09",
    "CUSTOM_10",
    "CUSTOM_11",
    "CUSTOM_12",
    "CUSTOM_13",
    "CUSTOM_14",
    "CUSTOM_15",
    "CUSTOM_16",
    "CUSTOM_17",
    "CUSTOM_18",
    "CUSTOM_19",
    "CUSTOM_20",
    "CUSTOM_21",
    "CUSTOM_22",
    "CUSTOM_23",
    "CUSTOM_24",
    "CUSTOM_25",
    "CUSTOM_26",
    "CUSTOM_27",
    "CUSTOM_28",
    "CUSTOM_29",
    "CUSTOM_30",
    "CUSTOM_31",
    "CUSTOM_32",
    "CUSTOM_33",
    "CUSTOM_34",
    "CUSTOM_35",
    "CUSTOM_36",
    "CUSTOM_37",
    "CUSTOM_38",
    "CUSTOM_39",
    "CUSTOM_40",
    "CUSTOM_41",
    "CUSTOM_42",
    "CUSTOM_43",
    "CUSTOM_44",
    "CUSTOM_45",
    "CUSTOM_46",
    "CUSTOM_47",
    "CUSTOM_48",
    "CUSTOM_49",
    "CUSTOM_50",
    "CUSTOM_51",
    "CUSTOM_52",
    "CUSTOM_53",
    "CUSTOM_54",
    "CUSTOM_55",
    "CUSTOM_56",
    "CUSTOM_57",
    "CUSTOM_58",
    "CUSTOM_59",
    "CUSTOM_60",
    "CUSTOM_61",
    "CUSTOM_62",
    "CUSTOM_63",
    "CUSTOM_64",
    "CUSTOM_65",
    "CUSTOM_66",
    "CUSTOM_67",
    "CUSTOM_68",
    "CUSTOM_69",
    "CUSTOM_70",
    "CUSTOM_71",
    "CUSTOM_72",
    "CUSTOM_73",
    "CUSTOM_74",
    "CUSTOM_75",
    "CUSTOM_76",
    "CUSTOM_77",
    "CUSTOM_78",
    "CUSTOM_79",
    "CUSTOM_80",
    "CUSTOM_81",
    "CUSTOM_82",
    "CUSTOM_83",
    "CUSTOM_84",
    "CUSTOM_85",
    "CUSTOM_86",
    "CUSTOM_87",
    "CUSTOM_88",
    "CUSTOM_89",
    "CUSTOM_90",
    "CUSTOM_91",
    "CUSTOM_92",
    "CUSTOM_93",
    "CUSTOM_94",
    "CUSTOM_95",
    "CUSTOM_96",
    "CUSTOM_97",
    "CUSTOM_98",
    "CUSTOM_99",
    "CUSTOM_100",
    "CUSTOM_101",
    "CUSTOM_102",
    "CUSTOM_103",
    "CUSTOM_104",
    "CUSTOM_105",
    "CUSTOM_106",
    "CUSTOM_107",
    "CUSTOM_108",
    "CUSTOM_109",
    "CUSTOM_110",
    "CUSTOM_111",
    "CUSTOM_112",
    "CUSTOM_113",
    "CUSTOM_114",
    "CUSTOM_115",
    "CUSTOM_116",
    "CUSTOM_117",
    "CUSTOM_118",
    "CUSTOM_119",
    "CUSTOM_120",
    "CUSTOM_121",
    "CUSTOM_122",
    "CUSTOM_123",
    "CUSTOM_124",
    "CUSTOM_125",
    "CUSTOM_126",
    "CUSTOM_127",
    "CUSTOM_128",
    "CUSTOM_129",
    "CUSTOM_130",
    "CUSTOM_131",
    "CUSTOM_132",
    "CUSTOM_133",
    "CUSTOM_134",
    "CUSTOM_135",
    "CUSTOM_136",
    "CUSTOM_137",
    "CUSTOM_138",
    "CUSTOM_139",
    "CUSTOM_140",
    "CUSTOM_141",
    "CUSTOM_142",
    "CUSTOM_143",
    "CUSTOM_144",
    "CUSTOM_145",
    "CUSTOM_146",
    "CUSTOM_147",
    "CUSTOM_148",
    "CUSTOM_149",
    "CUSTOM_150",
    "CUSTOM_151",
    "CUSTOM_152",
    "CUSTOM_153",
    "CUSTOM_154",
    "CUSTOM_155",
    "CUSTOM_156",
    "CUSTOM_157",
    "CUSTOM_158",
    "CUSTOM_159",
    "CUSTOM_160",
    "CUSTOM_161",
    "CUSTOM_162",
    "CUSTOM_163",
    "CUSTOM_164",
    "CUSTOM_165",
    "CUSTOM_166",
    "CUSTOM_167",
    "CUSTOM_168",
    "CUSTOM_169",
    "CUSTOM_170",
    "CUSTOM_171",
    "CUSTOM_172",
    "CUSTOM_173",
    "CUSTOM_174",
    "CUSTOM_175",
    "CUSTOM_176",
    "CUSTOM_177",
    "CUSTOM_178",
    "CUSTOM_179",
    "CUSTOM_180",
    "CUSTOM_181",
    "CUSTOM_182",
    "CUSTOM_183",
    "CUSTOM_184",
    "CUSTOM_185",
    "CUSTOM_186",
    "CUSTOM_187",
    "CUSTOM_188",
    "CUSTOM_189",
    "CUSTOM_190",
    "CUSTOM_191",
    "CUSTOM_192",
    "CUSTOM_193",
    "CUSTOM_194",
    "CUSTOM_195",
    "CUSTOM_196",
    "CUSTOM_197",
    "CUSTOM_198",
    "CUSTOM_199",
    "CUSTOM_200",
    "CUSTOM_201",
    "CUSTOM_202",
    "CUSTOM_203",
    "CUSTOM_204",
    "CUSTOM_205",
    "CUSTOM_206",
    "CUSTOM_207",
    "CUSTOM_208",
    "CUSTOM_209",
    "CUSTOM_210",
    "CUSTOM_211",
    "CUSTOM_212",
    "CUSTOM_213",
    "CUSTOM_214",
    "CUSTOM_215",
    "CUSTOM_216",
    "CUSTOM_217",
    "CUSTOM_218",
    "CUSTOM_219",
    "CUSTOM_220",
    "CUSTOM_221",
    "CUSTOM_222",
    "CUSTOM_223",
    "CUSTOM_224",
    "CUSTOM_225",
    "CUSTOM_226",
    "CUSTOM_227",
    "CUSTOM_228",
    "CUSTOM_229",
    "CUSTOM_230",
    "CUSTOM_231",
    "CUSTOM_232",
    "CUSTOM_233",
    "CUSTOM_234",
    "CUSTOM_235",
    "CUSTOM_236",
    "CUSTOM_237",
    "CUSTOM_238",
    "CUSTOM_239",
    "CUSTOM_240",
    "CUSTOM_241",
    "CUSTOM_242",
    "CUSTOM_243",
    "CUSTOM_244",
    "CUSTOM_245",
    "CUSTOM_246",
    "CUSTOM_247",
    "CUSTOM_248",
    "CUSTOM_249",
    "CUSTOM_250",
    "CUSTOM_251",
    "CUSTOM_252",
    "CUSTOM_253",
    "CUSTOM_254",
    "CUSTOM_255",
    "CUSTOM_256",
    "OTHER_DRUGS",
    "MARIJUANA",
    "OTHER_EDUCATION",
    "CONTINUING_EDUCATION_COLLEGES",
    "HISTORY",
    "K_12",
    "REFERENCE_SITES",
    "SCIENCE_AND_TECHNOLOGY",
    "OTHER_ENTERTAINMENT_AND_RECREATION",
    "ENTERTAINMENT",
    "TELEVISION_AND_MOVIES",
    "MUSIC",
    "STREAMING_MEDIA",
    "RADIO_STATIONS",
    "GAMBLING",
    "OTHER_GAMES",
    "SOCIAL_NETWORKING_GAMES",
    "OTHER_GOVERNMENT_AND_POLITICS",
    "GOVERNMENT",
    "POLITICS",
    "HEALTH",
    "OTHER_ILLEGAL_OR_QUESTIONABLE",
    "COPYRIGHT_INFRINGEMENT",
    "COMPUTER_HACKING",
    "QUESTIONABLE",
    "PROFANITY",
    "MATURE_HUMOR",
    "ANONYMIZER",
    "OTHER_INFORMATION_TECHNOLOGY",
    "TRANSLATORS",
    "IMAGE_HOST",
    "FILE_HOST",
    "SHAREWARE_DOWNLOAD",
    "WEB_BANNERS",
    "WEB_HOST",
    "WEB_SEARCH",
    "PORTALS",
    "SAFE_SEARCH_ENGINE",
    "CDN",
    "OSS_UPDATES",
    "DNS_OVER_HTTPS",
    "OTHER_INTERNET_COMMUNICATION",
    "INTERNET_SERVICES",
    "DISCUSSION_FORUMS",
    "ONLINE_CHAT",
    "EMAIL_HOST",
    "BLOG",
    "P2P_COMMUNICATION",
    "REMOTE_ACCESS",
    "WEB_CONFERENCING",
    "ZSPROXY_IPS",
    "JOB_SEARCH",
    "MILITANCY_HATE_AND_EXTREMISM",
    "OTHER_MISCELLANEOUS",
    "MISCELLANEOUS_OR_UNKNOWN",
    "NEWLY_REG_DOMAINS",
    "NON_CATEGORIZABLE",
    "NEWS_AND_MEDIA",
    "OTHER_RELIGION",
    "TRADITIONAL_RELIGION",
    "CULT",
    "ALT_NEW_AGE",
    "OTHER_SECURITY",
    "ADWARE_OR_SPYWARE",
    "ENCR_WEB_CONTENT",
    "MALICIOUS_TLD",
    "OTHER_SHOPPING_AND_AUCTIONS",
    "SPECIALIZED_SHOPPING",
    "REAL_ESTATE",
    "ONLINE_AUCTIONS",
    "OTHER_SOCIAL_AND_FAMILY_ISSUES",
    "SOCIAL_ISSUES",
    "FAMILY_ISSUES",
    "OTHER_SOCIETY_AND_LIFESTYLE",
    "ART_CULTURE",
    "ALTERNATE_LIFESTYLE",
    "HOBBIES_AND_LEISURE",
    "DINING_AND_RESTAURANT",
    "ALCOHOL_TOBACCO",
    "SOCIAL_NETWORKING",
    "SPECIAL_INTERESTS",
    "SPORTS",
    "TASTELESS",
    "TRAVEL",
    "USER_DEFINED",
    "VEHICLES",
    "VIOLENCE",
    "WEAPONS_AND_BOMBS",
]

super_categories = [
    "ENTERTAINMENT",
    "MUSIC",
    "OTHER_ENTERTAINMENT_AND_RECREATION",
    "RADIO_STATIONS",
    "STREAMING_MEDIA",
    "TELEVISION_AND_MOVIES",
    "NEWS_AND_MEDIA",
    "USER_DEFINED",
    "CLASSIFIEDS",
    "CORPORATE_MARKETING",
    "FINANCE",
    "TRADING_BROKERAGE_INSURANCE",
    "OTHER_BUSINESS_AND_ECONOMY",
    "PROFESSIONAL_SERVICES",
    "CONTINUING_EDUCATION_COLLEGES",
    "HISTORY",
    "K_12",
    "OTHER_EDUCATION",
    "REFERENCE_SITES",
    "SCIENCE_AND_TECHNOLOGY",
    "WEB_BANNERS",
    "CDN",
    "DNS_OVER_HTTPS",
    "FILE_HOST",
    "IMAGE_HOST",
    "OSS_UPDATES",
    "OTHER_INFORMATION_TECHNOLOGY",
    "PORTALS",
    "SAFE_SEARCH_ENGINE",
    "SHAREWARE_DOWNLOAD",
    "TRANSLATORS",
    "WEB_HOST",
    "WEB_SEARCH",
    "BLOG",
    "DISCUSSION_FORUMS",
    "INTERNET_SERVICES",
    "ONLINE_CHAT",
    "OTHER_INTERNET_COMMUNICATION",
    "P2P_COMMUNICATION",
    "REMOTE_ACCESS",
    "EMAIL_HOST",
    "WEB_CONFERENCING",
    "ZSPROXY_IPS",
    "JOB_SEARCH",
    "GOVERNMENT",
    "OTHER_GOVERNMENT_AND_POLITICS",
    "POLITICS",
    "MISCELLANEOUS_OR_UNKNOWN",
    "NEWLY_REG_DOMAINS",
    "NON_CATEGORIZABLE",
    "OTHER_MISCELLANEOUS",
    "TRAVEL",
    "VEHICLES",
    "ADULT_SEX_EDUCATION",
    "ADULT_THEMES",
    "K_12_SEX_EDUCATION",
    "LINGERIE_BIKINI",
    "NUDITY",
    "OTHER_ADULT_MATERIAL",
    "PORNOGRAPHY",
    "SEXUALITY",
    "SOCIAL_ADULT",
    "DRUGS",
    "MARIJUANA",
    "GAMBLING",
    "ANONYMIZER",
    "COMPUTER_HACKING",
    "COPYRIGHT_INFRINGEMENT",
    "MATURE_HUMOR",
    "OTHER_ILLEGAL_OR_QUESTIONABLE",
    "PROFANITY",
    "QUESTIONABLE",
    "MILITANCY_HATE_AND_EXTREMISM",
    "TASTELESS",
    "VIOLENCE",
    "WEAPONS_AND_BOMBS",
    "OTHER_GAMES",
    "SOCIAL_NETWORKING_GAMES",
    "HEALTH",
    "ALT_NEW_AGE",
    "CULT",
    "OTHER_RELIGION",
    "TRADITIONAL_RELIGION",
    "ONLINE_AUCTIONS",
    "SPECIALIZED_SHOPPING",
    "OTHER_SHOPPING_AND_AUCTIONS",
    "REAL_ESTATE",
    "FAMILY_ISSUES",
    "OTHER_SOCIAL_AND_FAMILY_ISSUES",
    "SOCIAL_ISSUES",
    "ALCOHOL_TOBACCO",
    "LIFESTYLE",
    "ART_CULTURE",
    "DINING_AND_RESTAURANT",
    "HOBBIES_AND_LEISURE",
    "OTHER_SOCIETY_AND_LIFESTYLE",
    "SOCIAL_NETWORKING",
    "SPECIAL_INTERESTS",
    "SPORTS",
    "ENCR_WEB_CONTENT",
    "MALICIOUS_TLD",
    "OTHER_SECURITY",
    "ADWARE_OR_SPYWARE",
]

valid_countries = [
    "ANY",
    "NONE",
    "COUNTRY_AD",
    "COUNTRY_AE",
    "COUNTRY_AF",
    "COUNTRY_AG",
    "COUNTRY_AI",
    "COUNTRY_AL",
    "COUNTRY_AM",
    "COUNTRY_AN",
    "COUNTRY_AO",
    "COUNTRY_AQ",
    "COUNTRY_AR",
    "COUNTRY_AS",
    "COUNTRY_AT",
    "COUNTRY_AU",
    "COUNTRY_AW",
    "COUNTRY_AZ",
    "COUNTRY_BA",
    "COUNTRY_BB",
    "COUNTRY_BD",
    "COUNTRY_BE",
    "COUNTRY_BF",
    "COUNTRY_BG",
    "COUNTRY_BH",
    "COUNTRY_BI",
    "COUNTRY_BJ",
    "COUNTRY_BM",
    "COUNTRY_BN",
    "COUNTRY_BO",
    "COUNTRY_BR",
    "COUNTRY_BS",
    "COUNTRY_BT",
    "COUNTRY_BV",
    "COUNTRY_BW",
    "COUNTRY_BY",
    "COUNTRY_BZ",
    "COUNTRY_CA",
    "COUNTRY_CC",
    "COUNTRY_CD",
    "COUNTRY_CF",
    "COUNTRY_CG",
    "COUNTRY_CH",
    "COUNTRY_CI",
    "COUNTRY_CK",
    "COUNTRY_CL",
    "COUNTRY_CM",
    "COUNTRY_CN",
    "COUNTRY_CO",
    "COUNTRY_CR",
    "COUNTRY_CU",
    "COUNTRY_CV",
    "COUNTRY_CX",
    "COUNTRY_CY",
    "COUNTRY_CZ",
    "COUNTRY_DE",
    "COUNTRY_DJ",
    "COUNTRY_DK",
    "COUNTRY_DM",
    "COUNTRY_DO",
    "COUNTRY_DZ",
    "COUNTRY_EC",
    "COUNTRY_EE",
    "COUNTRY_EG",
    "COUNTRY_EH",
    "COUNTRY_ER",
    "COUNTRY_ES",
    "COUNTRY_ET",
    "COUNTRY_FI",
    "COUNTRY_FJ",
    "COUNTRY_FK",
    "COUNTRY_FM",
    "COUNTRY_FO",
    "COUNTRY_FR",
    "COUNTRY_FX",
    "COUNTRY_GA",
    "COUNTRY_GB",
    "COUNTRY_GD",
    "COUNTRY_GE",
    "COUNTRY_GF",
    "COUNTRY_GH",
    "COUNTRY_GI",
    "COUNTRY_GL",
    "COUNTRY_GM",
    "COUNTRY_GN",
    "COUNTRY_GP",
    "COUNTRY_GQ",
    "COUNTRY_GR",
    "COUNTRY_GS",
    "COUNTRY_GT",
    "COUNTRY_GU",
    "COUNTRY_GW",
    "COUNTRY_GY",
    "COUNTRY_HK",
    "COUNTRY_HM",
    "COUNTRY_HN",
    "COUNTRY_HR",
    "COUNTRY_HT",
    "COUNTRY_HU",
    "COUNTRY_ID",
    "COUNTRY_IE",
    "COUNTRY_IL",
    "COUNTRY_IN",
    "COUNTRY_IO",
    "COUNTRY_IQ",
    "COUNTRY_IR",
    "COUNTRY_IS",
    "COUNTRY_IT",
    "COUNTRY_JM",
    "COUNTRY_JO",
    "COUNTRY_JP",
    "COUNTRY_KE",
    "COUNTRY_KG",
    "COUNTRY_KH",
    "COUNTRY_KI",
    "COUNTRY_KM",
    "COUNTRY_KN",
    "COUNTRY_KP",
    "COUNTRY_KR",
    "COUNTRY_KW",
    "COUNTRY_KY",
    "COUNTRY_KZ",
    "COUNTRY_LA",
    "COUNTRY_LB",
    "COUNTRY_LC",
    "COUNTRY_LI",
    "COUNTRY_LK",
    "COUNTRY_LR",
    "COUNTRY_LS",
    "COUNTRY_LT",
    "COUNTRY_LU",
    "COUNTRY_LV",
    "COUNTRY_LY",
    "COUNTRY_MA",
    "COUNTRY_MC",
    "COUNTRY_MD",
    "COUNTRY_MG",
    "COUNTRY_MH",
    "COUNTRY_MK",
    "COUNTRY_ML",
    "COUNTRY_MM",
    "COUNTRY_MN",
    "COUNTRY_MO",
    "COUNTRY_MP",
    "COUNTRY_MQ",
    "COUNTRY_MR",
    "COUNTRY_MS",
    "COUNTRY_MT",
    "COUNTRY_MU",
    "COUNTRY_MV",
    "COUNTRY_MW",
    "COUNTRY_MX",
    "COUNTRY_MY",
    "COUNTRY_MZ",
    "COUNTRY_NA",
    "COUNTRY_NC",
    "COUNTRY_NE",
    "COUNTRY_NF",
    "COUNTRY_NG",
    "COUNTRY_NI",
    "COUNTRY_NL",
    "COUNTRY_NO",
    "COUNTRY_NP",
    "COUNTRY_NR",
    "COUNTRY_NU",
    "COUNTRY_NZ",
    "COUNTRY_OM",
    "COUNTRY_PA",
    "COUNTRY_PE",
    "COUNTRY_PF",
    "COUNTRY_PG",
    "COUNTRY_PH",
    "COUNTRY_PK",
    "COUNTRY_PL",
    "COUNTRY_PM",
    "COUNTRY_PN",
    "COUNTRY_PR",
    "COUNTRY_PS",
    "COUNTRY_PT",
    "COUNTRY_PW",
    "COUNTRY_PY",
    "COUNTRY_QA",
    "COUNTRY_RE",
    "COUNTRY_RO",
    "COUNTRY_RU",
    "COUNTRY_RW",
    "COUNTRY_SA",
    "COUNTRY_SB",
    "COUNTRY_SC",
    "COUNTRY_SD",
    "COUNTRY_SE",
    "COUNTRY_SG",
    "COUNTRY_SH",
    "COUNTRY_SI",
    "COUNTRY_SJ",
    "COUNTRY_SK",
    "COUNTRY_SL",
    "COUNTRY_SM",
    "COUNTRY_SN",
    "COUNTRY_SO",
    "COUNTRY_SR",
    "COUNTRY_ST",
    "COUNTRY_SV",
    "COUNTRY_SY",
    "COUNTRY_SZ",
    "COUNTRY_TC",
    "COUNTRY_TD",
    "COUNTRY_TF",
    "COUNTRY_TG",
    "COUNTRY_TH",
    "COUNTRY_TJ",
    "COUNTRY_TK",
    "COUNTRY_TM",
    "COUNTRY_TN",
    "COUNTRY_TO",
    "COUNTRY_TL",
    "COUNTRY_TR",
    "COUNTRY_TT",
    "COUNTRY_TV",
    "COUNTRY_TW",
    "COUNTRY_TZ",
    "COUNTRY_UA",
    "COUNTRY_UG",
    "COUNTRY_UM",
    "COUNTRY_US",
    "COUNTRY_UY",
    "COUNTRY_UZ",
    "COUNTRY_VA",
    "COUNTRY_VC",
    "COUNTRY_VE",
    "COUNTRY_VG",
    "COUNTRY_VI",
    "COUNTRY_VN",
    "COUNTRY_VU",
    "COUNTRY_WF",
    "COUNTRY_WS",
    "COUNTRY_YE",
    "COUNTRY_YT",
    "COUNTRY_RS",
    "COUNTRY_ZA",
    "COUNTRY_ZM",
    "COUNTRY_ME",
    "COUNTRY_ZW",
    "COUNTRY_AX",
    "COUNTRY_GG",
    "COUNTRY_IM",
    "COUNTRY_JE",
    "COUNTRY_BL",
    "COUNTRY_MF",
]



import json
import pdb  # noqa
import time

import requests

from zscaler_api_talkers.helpers import HttpCalls, setup_logger
from zscaler_api_talkers.zia.models import (
    super_categories,
    valid_category_ids,
    valid_countries,
)

from zscaler_api_talkers.zia.helpers import _obfuscate_api_key

logger = setup_logger(name=__name__)


class ZiaTalker(object):
    """
    ZIA API talker
    Documentation:
    https://help.zscaler.com/zia/zia-api/api-developer-reference-guide
    """

    def __init__(
        self,
        cloud_name: str,
        bearer: str = None,
        api_key: str = "",
        username: str = "",
        password: str = "",
    ):
        """
        Method to start the class

        :param cloud_name: (str) Example: zscalerbeta.net, zscalerone.net, zscalertwo.net, zscalerthree.net,
            zscaler.net, zscloud.net
        :param bearer: (str) OAuth2.0 Bear token
        """
        self.base_uri = f"https://zsapi.{cloud_name}/api/v1"
        self.hp_http = HttpCalls(
            host=self.base_uri,
            verify=True,
        )
        self.cookies = None
        self.headers = None
        if bearer:
            self.headers = {"Authorization": f"Bearer {bearer}"}
        elif username and any([password, api_key]):
            self.authenticate(
                username=username,
                api_key=api_key,
                password=password,
            )

    def authenticate(
        self,
        api_key: str,
        username: str,
        password: str = None,
    ):
        """
        Method to authenticate.

        :param api_key: (str) API key
        :param username: (str) A string that contains the email ID of the API admin
        :param password: (str) A string that contains the password for the API admin
        """
        timestamp, key = _obfuscate_api_key(api_key)
        payload = {
            "apiKey": key,
            "username": username,
            "password": password,
            "timestamp": timestamp,
        }
        url = "/authenticatedSession"
        response = self.hp_http.post_call(
            url=url,
            payload=payload,
        )
        self.cookies = {"JSESSIONID": response.cookies["JSESSIONID"]}

    def authenticated_session(self) -> json:
        """
        Checks if there is an authenticated session

        :return: (json)
        """
        url = "/authenticatedSession"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
        )

        return response.json()

    def end_session(self) -> json:
        """
        Method to end an authenticated session

        :return: (json)
        """
        url = "/authenticatedSession"
        response = self.hp_http.delete_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            payload={},
        )

        return response.json()

    def _obtain_all(
        self,
        url: str,
    ) -> list:
        """
        Internal method that queries all pages

        :param url: (str) URL

        :return: (list) List of results
        """
        page = 1
        result = []
        while True:
            response = self.hp_http.get_call(
                url=f"{url}&page={page}",
                cookies=self.cookies,
                error_handling=True,
                headers=self.headers,
            )
            if response.json():
                result += response.json()
                page += 1
                time.sleep(1)
            else:
                break

        return result

    def get_status(self) -> json:
        """
        Method to obtain the activation status for a configuration change

        :return: (json) JSON object with the status
        """
        url = "/status"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
        )

        return response.json()

    def activate_status(self) -> json:
        """
        Method to activate configuration changes

        :return: (json) JSON object with the status
        """
        url = "/status/activate"
        response = self.hp_http.post_call(
            url,
            payload={},
            cookies=self.cookies,
            error_handling=True,
        )

        return response.json()

    # Admin Audit Logs

    def list_auditlog_entry_report(self) -> json:
        """
        Gets the status of a request for an audit log report.  After sending a POST request to /auditlogEntryReport
        to generate a report, you can continue to call GET /auditlogEntryReport to check whether the report has
        finished generating. Once the status is COMPLETE, you can send another GET request to
        /auditlogEntryReport/download to download the report as a CSV file.

        :return: (json)
        """
        url = "/auditlogEntryReport"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            headers=self.headers,
            error_handling=True,
        )

        return response.json()

    def download_auditlog_entry_report(self) -> requests.Response:
        """
        Gets the status of a request for an audit log report. After sending a POST request to /auditlogEntryReport
        to generate a report, you can continue to call GET /auditlogEntryReport to check whether the report has
        finished generating. Once the status is COMPLETE, you can send another GET request to
        /auditlogEntryReport/download to download the report as a CSV file.

        :return: (request.Response)
        """
        url = "/auditlogEntryReport/download"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            headers=self.headers,
            error_handling=True,
        )

        return response

    def add_auditlog_entry_report(
        self,
        start_time: int,
        end_time: int,
        action_types: list = None,
        category: str = None,
        subcategories: list = None,
        action_interface: str = None,
    ) -> requests.Response:
        """
         Creates an audit log report for the specified time period and saves it as a CSV file. The report includes
         audit information for every call made to the cloud service API during the specified time period. Creating a
         new audit log report will overwrite a previously-generated report.

        :param start_time: (int) The timestamp, in epoch, of the admin's last login
        :param end_time: (int) The timestamp, in epoch, of the admin's last logout.
        :param action_types: (list) The action performed by the admin in the ZIA Admin Portal or API
        :param category: (str) The location in the Zscaler Admin Portal (i.e., Admin UI) where the actionType was
        performed.
        :param subcategories: (list) The area within a category where the actionType was performed.
        :param action_interface: (str) The interface (i.e., Admin UI or API) where the actionType was performed.

        :return: (requests.Response Object) 204 Successful Operation
        """
        url = "/auditlogEntryReport"
        payload = {
            "startTime": start_time,
            "endTime": end_time,
        }
        if category:
            payload.update(category=category)
        if subcategories:
            payload.update(subcategories=subcategories)
        if action_interface:
            payload.update(actionInterface=action_interface)
        if action_types:
            payload.update(actionTypes=action_types)

        response = self.hp_http.post_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response

    # Admin & Role Management
    def list_admin_users(
        self,
        user_id: int = None,
        query: str = None,
    ) -> json:
        """
        Gets a list of admin users. By default, auditor user information is not included.

        :param user_id: (int) user ID
        :param query: (str) HTTP query  # TODO: What is this?  Looks like it is just parameters

        :return: (json) JSON of results
        """
        if user_id:
            url = f"/adminUsers/{user_id}"
            return self.hp_http.get_call(
                url,
                cookies=self.cookies,
                error_handling=True,
                headers=self.headers,
            ).json()
        else:
            if query:
                url = f"/adminUsers?{query}?pageSize=1000"
            else:
                url = "/adminUsers?pageSize=1000"

        return self._obtain_all(url)

    def add_admin_users(self, loginName: str, userName: str, email: str, password: str, role: dict, comments: str = '',
                       adminScopeType: str ='ORGANIZATION',
                       adminScopeScopeEntities: list =[],
                       adminScopescopeGroupMemberEntities: list =[],
                       isNonEditable: bool = False,
                       disabled: bool = False,
                       isAuditor: bool = False,
                       isPasswordLoginAllowed: object = False,
                       isSecurityReportCommEnabled: object = False,
                       isServiceUpdateCommEnabled: object = False,
                       isProductUpdateCommEnabled: object = False,
                       isPasswordExpired: object = False,
                       isExecMobileAppEnabled: object = False,
                       execMobileAppTokens: object = []) -> json:
        """
        Adds a new Admininstrator.
               :param loginName: string. Admin or auditor's login name. loginName is in email format
               and uses the domain name associated to the Zscaler account.
               :param userName: string. UserName.
               :param email: string. Email Address.
               :param password: string. Password for administrator. If admin single sign-on (SSO) is disabled, then this field is mandatory
               :param role : Role of the Admin
               :param comments: string. Comments.
               :param adminScopeType: string. Scope of the admin.
               :param adminScopeScopeEntities: list: Department or Location when adminScopeType is set to Deportment or Location.
               :param adminScopescopeGroupMemberEntities: list. Location Groups when adminScopeType is set to Location Group.
               :param isNonEditable: boolean. Indicates whether or not the admin can be edited or deleted. default: False.
               :param disabled: boolean. If admin accounts is disabled. default: False.
               :param isAuditor:boolean. Indicates if user is auditor. default: False.
               :param isPasswordLoginAllowed: boolean. If password login is allowed. default: False.
               :param isSecurityReportCommEnabled: boolean. Communication for Security Report is enabled. default: False.
               :param isServiceUpdateCommEnabled: boolean. Communication setting for Service Update. default: False.
               :param isProductUpdateCommEnabled: boolean. Communication setting for Product Update. default: False.
               :param isPasswordExpired: boolean. Expire password to force user to change password on logon. default: False.
               :param isExecMobileAppEnabled: boolean. Indicates whether or not Executive Insights App access is enabled for the admin. default: False.
               :return:json()
               """
        url = "/adminUsers"
        payload = {
            "loginName": loginName,
            "userName": userName,
            "email": email,
            "password": password,
            "role": role,
            "comments": comments,
            "adminScopeType": adminScopeType,
            "adminScopeScopeEntities": adminScopeScopeEntities,
            "adminScopescopeGroupMemberEntities": adminScopescopeGroupMemberEntities,
            "isNonEditable": isNonEditable,
            "disabled": disabled,
            "isAuditor": isAuditor,
            "isPasswordLoginAllowed": isPasswordLoginAllowed,
            "isSecurityReportCommEnabled": isSecurityReportCommEnabled,
            "isServiceUpdateCommEnabled": isServiceUpdateCommEnabled,
            "isProductUpdateCommEnabled": isProductUpdateCommEnabled,
            "isPasswordExpired": isPasswordExpired,
            "isExecMobileAppEnabled": isExecMobileAppEnabled,
            "execMobileAppTokens": execMobileAppTokens
        }
        response = self.hp_http.post_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )
        return response.json()
    def list_admin_roles(
        self,
        query: str = None,
    ) -> json:
        """
        Gets a name and ID dictionary of al admin roles

        :param query: (str) HTTP query  # TODO: What is this?  Looks like it is just parameters

        :return: (json)
        """
        if query:
            url = f"/adminRoles/lite?{query}"
        else:
            url = "/adminRoles/lite"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    # URL Categories
    def list_url_categories(
        self,
        custom: bool = False,
    ) -> json:
        """
        Gets information about all or custom URL categories

        :param custom: (bool) If True it will return custom categories only.  Default is False.

        :return: (json)
        """

        if custom:
            url = "/urlCategories?customOnly=true"
        else:
            url = "/urlCategories"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_url_categories_lite(self) -> json:
        """
        Gets a lightweight key-value list of all or custom URL categories.

        :return: (json)
        """
        url = "/urlCategories/lite"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def add_url_categories(
        self,
        name: str,
        super_category: str,
        type_list: str = None,
        urls: list = None,
        db_categorized_urls: list = None,
        keywords_retaining_parent_category: list = None,
        keywords: list = None,
        custom_category: bool = False,
        ip_ranges: list = None,
        ip_ranges_retaining_parent_category: list = None,
        description: str = None,
    ) -> json:
        """
         Adds a new custom URL category.

        :param name: (str) Name of the custom category. Possible values URL_CATEGORY, TLD_CATEGORY, ALL
        :param super_category: (str) super category
        :param type_list: (list)
        :param urls: (list) List of urls
        :param db_categorized_urls: (list) URL retaining parent category
        :param keywords_retaining_parent_category: (list) Retained custom keywords from the parent URL category that is
        associated to a URL category.
        :param keywords: (list) Custom keywords associated to a URL category.
        :param custom_category: (bool) Default False. Set to True for custom category
        :param ip_ranges: (list) Custom IP address ranges associated to a URL category
        :param ip_ranges_retaining_parent_category: (list) The retaining parent custom IP address ranges associated to a
        URL category.
        :param description: (str) Description or notes

        :return:  json
        """
        if not type_list:
            type_list = "URL_CATEGORY"

        if keywords_retaining_parent_category is None:
            keywords_retaining_parent_category = []

        if super_category not in super_categories:
            logger.error(f"Invalid Super Category: {super_categories}")
            raise ValueError("Invalid super category")

        if keywords is None:
            keywords = []

        if ip_ranges is None:
            ip_ranges = []

        url = "/urlCategories"
        payload = {
            "configuredName": name,
            "customCategory": custom_category,
            "superCategory": super_category,
            "keywordsRetainingParentCategory": keywords_retaining_parent_category,
            "keywords": keywords,
            "urls": urls,
            "dbCategorizedUrls": db_categorized_urls,
            "ipRanges": ip_ranges,
            "ipRangesRetainingParentCategory": ip_ranges_retaining_parent_category,
            "type": type_list,
            "description": description,
        }
        response = self.hp_http.post_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def add_raw_url_categories(
        self,
        payload: dict,
    ) -> json:
        """
         Adds a new custom URL category.

        :param payload: (dict)

        :return: (json)
        """
        url = "/urlCategories"
        response = self.hp_http.post_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def update_url_categories(
        self,
        category_id: str,
        action: str = None,
        configured_name: str = None,
        urls: list = None,
        db_categorized_urls: list = None,
        keywords: list = None,
        keywords_retaining_parent_category: list = None,
    ) -> json:
        """
        Updates the URL category for the specified ID. If keywords are included within the request, then they will
        replace existing ones for the specified URL category . If the keywords attribute is not included the request,
        the existing keywords are retained. You can perform a full update for the specified URL category. However,
        if attributes are omitted within the update request then clear the values for those attributes.

        You can also perform an incremental update, to add or remove URLs for the specified URL category using the
        action parameter.

        :param category_id: (str) URL id
        :param action: (str) Optional parameter. ADD_TO_LIST or REMOVE_FROM_LIST
        :param configured_name: (str) Name of the custom category
        :param urls: (list) List of urls
        :param db_categorized_urls: (list) URL retaining parent category
        :param keywords: (list)
        :param keywords_retaining_parent_category: (list) List of key works

        :return:  (json)
        """
        """if categoryId not in valid_category_ids:
            print(f'Error -> Invalid category id')
            raise ValueError("Invalid category id")"""
        url = f"/urlCategories/{category_id}"
        parameters = {}

        if action and action not in ["ADD_TO_LIST", "REMOVE_FROM_LIST"]:
            logger.error(f"Invalid action: {action}")
            raise ValueError("Invalid action")
        else:
            parameters.update({"action": action})

        payload = {
            "configuredName": configured_name,
        }
        if keywords_retaining_parent_category:
            payload.update(
                keywordsRetainingParentCategory=keywords_retaining_parent_category
            )
        if keywords:
            payload.update(keywords=keywords)
        if configured_name:
            payload.update(configuredName=configured_name)
        if urls:
            payload.update(urls=urls)
        if db_categorized_urls:
            payload.update(dbCategorizedUrls=db_categorized_urls)

        response = self.hp_http.put_call(
            url,
            params=parameters,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def delete_url_categories(
        self,
        category_id: str,
    ) -> requests.Response:
        """
        Deletes the custom URL category for the specified ID. You cannot delete a custom category while it is being
        used by a URL policy or NSS feed. Also, predefined categories cannot be deleted.

        :param category_id: (inst) Category ID

        :return: (requests.Response)
        """
        url = f"/urlCategories/{category_id}"
        response = self.hp_http.delete_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response

    def delete_url_filtering_rules(
        self,
        rule_id: int,
    ) -> requests.Response:
        """
        Deletes the custom URL category for the specified ID. You cannot delete a custom category while it is being
        used by a URL policy or NSS feed. Also, predefined categories cannot be deleted.

        :param rule_id: (int) Rule Id

        :return: (request.Response)
        """
        url = f"/urlFilteringRules/{rule_id}"
        response = self.hp_http.delete_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response

    def list_url_categories_url_quota(self) -> json:
        """
        Gets information on the number of unique URLs that are currently provisioned for your organization as well as
        how many URLs you can add before reaching that number.

        :return: (json)
        """
        url = "/urlCategories/urlQuota"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_url_categories_id(
        self,
        category_id: int,
    ) -> json:
        """
        Gets the URL category information for the specified ID

        :param category_id: (int)

        :return: (json)
        """
        url = f"/urlCategories/{category_id}"
        if category_id not in valid_category_ids:
            logger.error(f"Invalid Category ID: {category_id}")
            raise ValueError("Invalid Category ID")
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def url_lookup(
        self,
        url_list: list,
    ) -> list:
        """
        Method to look up the categorization of the given list of URLs, ["abc.com","zyz.com"]

        :param url_list: (list) List of urls

        :return: (list)
        """
        result = []
        url = "/urlLookup"
        # Verify urls format
        list(set(url_list))
        # Rate limit 1/sec  and 400 hr and 100 URLs per call
        list_of_lists = [url_list[i : i + 100] for i in range(0, len(url_list), 100)]
        for item in list_of_lists:
            response = self.hp_http.post_call(
                url,
                payload=item,
                cookies=self.cookies,
                headers=self.headers,
                error_handling=True,
            )
            result.append(response.json())
            time.sleep(1)
        final_result = []
        for i in result:
            for j in i:
                final_result.append(j)

        return final_result

    # URL filtering Policies
    def list_url_filtering_rules(self) -> json:
        """
        Gets a list of all of URL Filtering Policy rules

        :return: (json)
        """
        url = "/urlFilteringRules"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def add_url_filtering_rules(  # FIXME: docstring lists params that aren't options and some params I don't know what their typehint should be.
        self,
        name: str,
        order: int,
        protocols: str,
        state: str,
        action: str,
        url_categories: list = None,
        request_methods: list = None,
        description=None,
        groups: list = None,
        locations: list = None,
        departments: list = None,
        users: list = None,
        rank: int = 7,
        location_groups=None,
        enforce_time_validity: bool = False,
        validity_end_time=None,
        validity_start_time=None,
        validity_time_zone_id=None,
        cbi_profile_id: int = 0,
        block_override: bool = False,
        **kwargs,
    ) -> json:
        """
         Adds a URL Filtering Policy rule. If you are using the Rank feature, refer to About Admin Rank to
         determine which value to provide for rank when adding a policy rule. If you are not using Admin Rank,
         the rank value must be 7.

        :param name: (str)  Name of the rule
        :param order: (int) Rule order
        :param protocols: (str) Possible values: SMRULEF_ZPA_BROKERS_RULE, ANY_RULE, TCP_RULE, UDP_RULE,
        DOHTTPS_RULE, TUNNELSSL_RULE, HTTP_PROXY, FOHTTP_RULE, FTP_RULE, HTTPS_RULE, HTTP_RULE, SSL_RULE, TUNNEL_RULE
        :param state: (str) enabled/disabled
        :param action: (str) Allow, Caution, Block
        :param url_categories: (list) List of URL categories for which rule must be applied
        :param request_methods: (list) Request method for which the rule must be applied. If not set, rule will be
        applied to all methods
        :param description: (str) Additional information about the URL Filtering rule
        :param groups: (list) Name-ID pairs of groups for which rule must be applied
        :param locations: (list) Each element is a dictionary: Name-ID pairs of locations for which rule must be applied
        :param departments: (list) Name-ID pairs of departments for which rule will be applied
        :param users: (list) Name-ID pairs of users for which rule must be applied
        :param rank: (int) Admin rank of the admin who creates this rule
        :param location_groups:
        :param enforce_time_validity: (bool)
        :param validity_end_time:
        :param validity_start_time:
        :param validity_time_zone_id:
        :param cbi_profile_id: (int)
        :param block_override: (bool) When set to true, a 'BLOCK' action triggered by the rule could be overridden.
        If true and both overrideGroup and overrideUsers are not set, the BLOCK triggered by this rule could be
        overridden for any users. If blockOverride is not set, 'BLOCK' action cannot be overridden.

        :param timewindows: (list) Name-ID pairs of time interval during which rule must be enforced.
        :param endUserNotificationUrl: (str) URL of end user notification page to be displayed when the rule
        is matched. Not applicable if either 'overrideUsers' or 'overrideGroups' is specified.
        :param overrideusers: Name-ID pairs of users for which this rule can be overridden. Applicable only if
        blockOverride is set to 'true', action is 'BLOCK' and overrideGroups is not set.If this overrideUsers is not
        set, 'BLOCK' action can be overridden for any user.
        :param overridegroups: Name-ID pairs of groups for which this rule can be overridden. Applicable only if
        blockOverride is set to 'true' and action is 'BLOCK'. If this overrideGroups is not set, 'BLOCK' action can
        be overridden for any group

        :return:
        """
        url = "/urlFilteringRules"
        payload = {
            "blockOverride": block_override,
            "cbiProfileId": cbi_profile_id,
            "description": description,
            "enforceTimeValidity": enforce_time_validity,
            "name": name,
            "order": order,
            "protocols": protocols,
            "urlCategories": url_categories,
            "state": state,
            "rank": rank,
            "action": action,
        }
        payload.update(kwargs)
        if locations:
            payload.update(locations=locations)
        if location_groups:
            payload.update(locationGroups=location_groups)
        if groups:
            payload.update(groups=groups)
        if departments:
            payload.update(departments=departments)
        if users:
            payload.update(users=users)
        if request_methods:
            payload.update(requestMethods=request_methods)
        if enforce_time_validity:
            payload.update(validityStartTime=validity_start_time)
            payload.update(validityEndTime=validity_end_time)
            payload.update(validityTimeZoneId=validity_time_zone_id)

        response = self.hp_http.post_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def update_url_filtering_rules(
        self,
        rule_id: int,
        **kwargs,
    ) -> json:
        url = f"/urlFilteringRules/{rule_id}"
        payload = kwargs
        response = self.hp_http.put_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    # User Management

    def list_departments(
        self,
        department_id: int = None,
    ) -> json or list:
        """
        Gets a list of departments. The search parameters find matching values within the "name" or "comments"
        attributes. if ID, gets the department for the specified ID

        :param department_id: (int) department ID

        :return: (json or list)
        """

        if not department_id:
            url = "/departments?pageSize=10000"
            return self._obtain_all(url)
        else:
            url = f"/departments/{department_id}"
            response = self.hp_http.get_call(
                url,
                cookies=self.cookies,
                error_handling=True,
                headers=self.headers,
            )
            return response.json()

    def list_groups(
        self,
        group_id: int = None,
    ) -> json or list:
        """
        Gets a list of groups if ID, gets the group for the specified ID

        :param group_id: group ID

        :return: (json)
        """
        if not group_id:
            url = "/groups?pageSize=10000"
            return self._obtain_all(url)
        else:
            url = f"/groups/{group_id}"
            response = self.hp_http.get_call(
                url,
                cookies=self.cookies,
                error_handling=True,
                headers=self.headers,
            )
            return response.json()

    def list_users(
        self,
        user_id: int = None,
        query: str = None,
    ) -> json or list:
        """
        Gets a list of all users and allows user filtering by name, department, or group. The name search parameter
        performs a partial match. The dept and group parameters perform a 'starts with' match. if ID,
        gets user information for the specified ID

        :param user_id: (int) user ID
        :param query: (str)

        :return: (json or list)
        """
        url = "/users?pageSize=1000"
        if user_id:
            url = f"/users/{user_id}"
            return self.hp_http.get_call(
                url,
                cookies=self.cookies,
                error_handling=True,
                headers=self.headers,
            ).json()
        elif query:
            url = f"/users?{query}&pageSize=1000"
            return self.hp_http.get_call(
                url,
                cookies=self.cookies,
                error_handling=True,
                headers=self.headers,
            ).json()

        return self._obtain_all(url)

    def add_users(
        self,
        name: str,
        email: str,
        groups: list,
        department: dict,
        comments: str,
        password: str,
        admin_user: bool = False,
    ) -> json:
        """
        Adds a new user. A user can belong to multiple groups, but can only belong to one department.

        :param name: (str) user name
        :param email: (str) user email address
        :param groups: (list) List each member is a dictionary, key id, value name [{"id":1234, "name":"guest-wifi"}]
        :param department: (dict) key is the id and value is the name {"id":1234, "name":"guests"}
        :param comments: (str) Comments
        :param password: (str) Password,
        :param admin_user: (bool) True if user is admin user. default False

        :return: (json)
        """
        url = "/users"
        payload = {
            "name": name,
            "email": email,
            "groups": groups,
            "department": department,
            "comments": comments,
            "password": password,
            "adminUser": admin_user,
        }
        response = self.hp_http.post_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def delete_bulk_users(
        self,
        user_ids: list,
    ) -> json:
        """
        Bulk delete users up to a maximum of 500 users per request. The response returns the user IDs that were
        successfully deleted.

        :param user_ids: (list) List of user IDS to be deleted.  Max 500 per bulk delete.

        :return: (json)
        """
        url = "/users/bulkDelete"
        if len(user_ids) < 500:
            payload = {"ids": user_ids}
            response = self.hp_http.post_call(
                url,
                payload=payload,
                cookies=self.cookies,
                error_handling=True,
                headers=self.headers,
            )
            return response.json()
        else:
            raise ValueError("Maximum 500 users per request")

    # Location Management

    def list_locations(
        self,
        location_id: int = None,
    ) -> json:
        """
        Gets locations only, not sub-locations. When a location matches the given search parameter criteria only its
        parent location is included in the result set, not its sub-locations.

        :param location_id: (int) Location id

        :return: (json)
        """
        url = "/locations"
        if location_id:
            url = f"/locations/{location_id}"

        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_sublocations(
        self,
        location_id: int,
    ) -> json:
        """
        Gets the sub-location information for the location with the specified ID

        :param location_id: (int) Location id

        :return: (json)
        """
        url = "/locations"
        if location_id:
            url = f"/locations/{location_id}/sublocations"

        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_locations_groups(self) -> json:
        """
        Gets information on location groups

        :return: (json)
        """
        url = "/locations/groups"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def delete_bulk_locations(
        self,
        location_ids: list,
    ) -> json:
        """
        Bulk delete locations up to a maximum of 100 users per request. The response returns the location IDs that
        were successfully deleted.

        :param location_ids: (list) List of location IDs

        :return: (json)
        """
        url = "/locations/bulkDelete"
        if len(location_ids) < 100:
            payload = {"ids": location_ids}
            response = self.hp_http.post_call(
                url,
                payload=payload,
                cookies=self.cookies,
                error_handling=True,
                headers=self.headers,
            )
            return response.json()
        else:
            raise ValueError("Maximum 100 locations per request")

    def delete_locations(
        self,
        location_id: int,
    ) -> requests.Response:
        """
        Deletes the location or sub-location for the specified ID

        :param location_id: (int) location ID

        :return: (request.Response object)
        """
        url = f"/locations/{location_id}"
        response = self.hp_http.delete_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response

    #   Traffic Forwarding

    def list_gre_tunnels(
        self,
        gre_tunnel_id: int = None,
    ) -> json:
        """
        Gets the GRE tunnel information for the specified ID

        :param gre_tunnel_id: (int) Optional. The unique identifier for the GRE tunnel

        :return: (json)
        """
        url = "/greTunnels"
        if gre_tunnel_id:
            url = f"/greTunnels/{gre_tunnel_id}"

        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def add_gre_tunnels(
        self,
        source_ip: str,
        primary_dest_vip: dict,
        secondary_dest_vip: dict,
        internal_ip_range: str,
        within_country: bool,
        comment: str,
        ip_unnumbered: bool,
    ) -> json:
        """
        Adds a GRE tunnel configuration.

        :param source_ip: (str) The source IP address of the GRE tunnel. This is typically a static IP address in the
        organization or SD-WAN. This IP address must be provisioned within the Zscaler service using the /staticIP
        endpoint.
        :param primary_dest_vip: (dict) {id:value} where value is integer: Unique identifier of the GRE primary VIP
        :param secondary_dest_vip: (dict) {id:value} where value is integer: Unique identifier of the GRE secondary VIP
        :param internal_ip_range: (str) The start of the internal IP address in /29 CIDR range
        :param within_country: (bool) Restrict the data center virtual IP addresses (VIPs) only to those within the
        same country as the source IP address
        :param comment: (str) Additional information about this GRE tunnel
        :param ip_unnumbered: (bool?) This is required to support the automated SD-WAN provisioning of GRE tunnels,
        when set to True gre_tun_ip and gre_tun_id are set to null

        :return: (json)
        """
        url = "/greTunnels"
        payload = {
            "sourceIp": source_ip,
            "primaryDestVip": primary_dest_vip,
            "secondaryDestVip": secondary_dest_vip,
            "internalIpRange": internal_ip_range,
            "withinCountry": within_country,
            "comment": comment,
            "ipUnnumbered": ip_unnumbered,
        }
        response = self.hp_http.post_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_gre_validate_and_get_available_internal_ip_ranges(self) -> json:
        """
        Gets the next available GRE tunnel internal IP address ranges

        :return: (json) List of available IP addresses
        """
        url = "/greTunnels/availableInternalIpRanges"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_gre_recommended_vips(
        self,
        query: str,
    ) -> json:
        """
        Gets a list of recommended GRE tunnel virtual IP addresses (VIPs), based on source IP address or
        latitude/longitude coordinates.

        :param query: (str) URL query. Example:

        :return: (json) List of available IP addresses
        """
        url = f"/vips/recommendedList?{query}"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_gre_validate_ip(
        self,
        ip: str,
    ) -> json:
        """
        Gets the static IP address and location mapping information for the specified GRE tunnel IP address

        :param ip: (str) IP address of the GRE tunnel.

        :return: (json)
        """
        url = f"/greTunnels/validateIP/{ip}"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_vpn_credentials(
        self,
        vpn_id: int = None,
    ) -> json:
        """
        Gets VPN credentials that can be associated to locations.

        :param vpn_id: (int) Optional. If specified, get VPN credentials for the specified ID.
        """
        url = "/vpnCredentials"
        if vpn_id:
            url = f"/vpnCredentials/{vpn_id}"

        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def add_vpn_credentials(
        self,
        fqdn: str,
        pre_shared_key: str,
        auth_type: str = "UFQDN",
        comments: str = None,
    ) -> json:
        """
        Adds VPN credentials that can be associated to locations.

        :param fqdn: (str) Example abc@domain.com
        :param pre_shared_key: (str) Pre-shared key. This is a required field for UFQDN and IP auth type
        :param auth_type: (str) VPN authentication type.
        valid options CN, IP, UFQDN,XAUTH
        :param comments: (str) Additional information about this VPN credential.

        :return: (json)
        """
        url = "/vpnCredentials"
        payload = {
            "type": auth_type,
            "fqdn": fqdn,
            "preSharedKey": pre_shared_key,
        }
        if comments:
            payload.update(comments=comments)
        response = self.hp_http.post_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def delete_vpn_credentials(
        self,
        vpn_id: int,
    ) -> requests.Response:  # TODO: Move to returning json
        """
        Deletes the VPN credentials for the specified ID.

        :param vpn_id: (int) The unique identifier for the VPN credential.

        :return: (requests.Response object)
        """
        url = f"/vpnCredentials/{vpn_id}"
        response = self.hp_http.delete_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response

    def list_static_ip(
        self,
        ip_id: int = None,
    ) -> json:
        """
        Gets all provisioned static IP addresses.

        :param ip_id: (str) Optional. If specified, get IP address for the specified id

        :return: (json)
        """
        url = "/staticIP"
        if ip_id:
            url = f"/staticIP/{ip_id}"

        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def add_static_ip(
        self,
        ip_address: str,
        geo_override: bool = False,
        routable_ip: bool = True,
        latitude: float = 0,
        longitude: float = 0,
        comment: str = "",
    ) -> json:
        """
        Adds a static IP address

        :param ip_address: (str) The static IP address
        :param geo_override: (bool) If not set, geographic coordinates and city are automatically determined from the
        IP address. Otherwise, the latitude and longitude coordinates must be provided.
        :param routable_ip: (bool) Indicates whether a non-RFC 1918 IP address is publicly routable. This attribute is
        ignored if there is no ZIA Private Service Edge associated to the organization.
        :param latitude: (float) Required only if the geoOverride attribute is set. Latitude with 7 digit precision
        after decimal point, ranges between -90 and 90 degrees.
        :param longitude: (float) Required only if the geoOverride attribute is set. Longitude with 7 digit precision
        after decimal point, ranges between -180 and 180 degrees.
        :param comment: (str) Additional information about this static IP address

        :return: (json)
        """
        url = "/staticIP"
        payload = {
            "ipAddress": ip_address,
            "latitude": latitude,
            "longitude": longitude,
            "routableIP": routable_ip,
            "comment": comment,
        }
        if geo_override:
            payload.update(geoOverrride=geo_override)
            payload.update(latitude=latitude)
            payload.update(longitude=longitude)

        response = self.hp_http.post_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def delete_static_ip(
        self,
        ip_address_id: int,
    ) -> requests.Response:
        """
        Deletes the static IP address for the specified ID.

        :param ip_address_id: (int) The unique identifier for the provisioned static IP address.

        :return: (request.Response object))
        """
        url = f"/staticIP/{ip_address_id}"
        response = self.hp_http.delete_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response

    # User Authentication Settings
    def list_exempted_urls(self) -> json:
        """
        Gets a list of URLs that were exempted from cookie authentication

        :return: (json)
        """
        url = "/authSettings/exemptedUrls"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def add_exempted_urls(
        self,
        urls: list,
    ) -> json:
        """
        Adds URLs to the cookie authentication exempt list to the list

        :param urls: (list) List of urls. Example ['url1','url2']

        :return: (json)
        """
        url = "/authSettings/exemptedUrls?action=ADD_TO_LIST"
        payload = {"urls": urls}
        response = self.hp_http.post_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def delete_exempted_urls(
        self,
        urls: list,
    ) -> json:
        """
        Removed URLs to the cookie authentication exempt list to the list

        :param urls: (list) List of urls. Example ['url1','url2']

        :return: (json)
        """
        url = "/authSettings/exemptedUrls"
        parameters = {"action": "REMOVE_FROM_LIST"}
        payload = {"urls": urls}
        response = self.hp_http.post_call(
            url,
            params=parameters,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    # Security Policy Settings

    def list_security_whitelisted_urls(self) -> json:
        """
        Gets a list of white-listed URLs

        :return: (json)
        """
        url = "/security"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def update_security_whitelisted_urls(
        self,
        urls: list,
    ) -> json:
        """
        Updates the list of white-listed URLs. This will overwrite a previously-generated white list. If you need to
        completely erase the white list, submit an empty list.

        :param urls: (list) List of urls ['www.zscaler.com', 'www.example.com']

        :return: (json)
        """
        url = "/security"
        payload = {"whitelistUrls": urls}
        response = self.hp_http.put_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_security_blacklisted_urls(self) -> json:
        """
        Gets a list of white-listed URLs

        :return: (json)
        """
        url = "/security/advanced"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def update_security_blacklisted_urls(
        self,
        urls: list,
    ) -> json:
        """
        Updates the list of black-listed URLs. This will overwrite a previously-generated black list. If you need to
        completely erase the black list, submit an empty list.

        :param urls: (list)

        :return: (json)
        """
        url = "/security/advanced"
        payload = {"blacklistUrls": urls}
        response = self.hp_http.put_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def add_security_blacklist_urls(
        self,
        urls: list,
    ) -> requests.Response:  # TODO: Move to return json
        """
        Adds a URL from the black list. To add a URL to the black list.

        :param urls: (list) List of urls

        :return: (request.Response object)
        """
        url = "/security/advanced/blacklistUrls"
        parameters = {"action": "ADD_TO_LIST"}
        payload = {"blacklistUrls": urls}
        response = self.hp_http.post_call(
            url,
            payload=payload,
            params=parameters,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response

    def remove_security_blacklist_urls(
        self,
        urls: list,
    ) -> json:
        """
        Removes a URL from the black list.

        :param urls: (list) List of urls

        :return: (json)
        """
        url = "/security/advanced/blacklistUrls"
        parameters = {"action": "REMOVE_FROM_LIST"}
        payload = {"blacklistUrls": urls}
        response = self.hp_http.post_call(
            url,
            payload=payload,
            params=parameters,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    # DLP Policies

    def list_dlp_dictionaries(
        self,
        dlp_dic_id: int = None,
    ) -> json:
        """
        Gets a list of all DLP Dictionaries.

        :param dlp_dic_id: (int) dlp dictionary id ( optional parameter)

        :return: (json)
        """
        url = "/dlpDictionaries"
        if dlp_dic_id:
            url = f"/dlpDictionaries/{dlp_dic_id}"

        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_dlp_dictionaries_lite(self) -> json:
        """
        Gets a list of all DLP Dictionary names and ID's only. T

        :return: (json)
        """
        url = "/dlpDictionaries/lite"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def validate_dlp_pattern(
        self,
        pattern: str,
    ) -> json:
        """
        Validates the pattern used by a Pattern and Phrases DLP dictionary type, and provides error information if
        the pattern is invalid.

        :param pattern: (str) Regex pattern

        :return: (json)
        """
        payload = pattern
        url = "/dlpDictionaries/validateDlpPattern"
        response = self.hp_http.post_call(
            url,
            cookies=self.cookies,
            payload=payload,  # TODO: payload is typically dict but here it is str?
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def delete_dlp_dictionaries(
        self,
        dlp_dic_id: int,
    ) -> requests.Response:
        """
        Deletes the custom DLP category for the specified ID. You cannot delete predefined DLP dictionaries. You
        cannot delete a custom dictionary while it is being used by a DLP Engine or policy. Also, predefined DLP
        dictionaries cannot be deleted.

        :param dlp_dic_id: (int) dlp dictionary ID

        :return: (requests.Response object)
        """
        url = f"/dlpDictionaries/{dlp_dic_id}"
        response = self.hp_http.delete_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response

    def add_dlp_dictionaries(
        self,
        dlp_dic_name: str,
        custom_phrase_match_type: str = "MATCH_ANY_CUSTOM_PHRASE_PATTERN_DICTIONARY",
        description: str = None,
        phrases: list = None,
        patterns: list = None,
    ) -> json:
        """
        Adds a new custom DLP dictionary that uses either Patterns and/or Phrases.

        :param dlp_dic_name: (str) Name
        :param custom_phrase_match_type: (str) customPhraseMatchType
        :param description: (str) description
        :param phrases: (list) list of phrases. valid example:[
        {"action": "PHRASE_COUNT_TYPE_UNIQUE", "phrase": "string"},
        {"action": "PHRASE_COUNT_TYPE_UNIQUE", "phrase": "string"}
        ]
        :param patterns: (list) list of patterns. valid example:[
        {"action": "PATTERN_COUNT_TYPE_UNIQUE", "phrase": "string"},
        {"action": "PATTERN_COUNT_TYPE_UNIQUE", "phrase": "string"}
        ]

        :return: (json)
        """
        if phrases is not None:
            for i in phrases:
                if i["action"] not in [
                    "PHRASE_COUNT_TYPE_UNIQUE",
                    "PHRASE_COUNT_TYPE_ALL",
                ]:
                    raise ValueError("Invalid action")
        if patterns is not None:
            for k in patterns:
                if k["action"] not in [
                    "PATTERN_COUNT_TYPE_UNIQUE",
                    "PATTERN_COUNT_TYPE_ALL",
                ]:
                    raise ValueError("Invalid action")

        if custom_phrase_match_type not in [
            "MATCH_ALL_CUSTOM_PHRASE_PATTERN_DICTIONARY",
            "MATCH_ANY_CUSTOM_PHRASE_PATTERN_DICTIONARY",
        ]:
            raise ValueError("Invalid customPhraseMatchType")

        url = "/dlpDictionaries"
        payload = {
            "name": dlp_dic_name,
            "description": description,
            "confidenceThreshold": None,
            "customPhraseMatchType": custom_phrase_match_type,
            "dictionaryType": "PATTERNS_AND_PHRASES",
            "phrases": phrases,
            "patterns": patterns,
        }
        response = self.hp_http.post_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_dlp_engines(
        self,
        dlp_engine_id: int = None,
    ) -> json:
        """
        Get a list of DLP engines.

        :param dlp_engine_id: (int) Optional value. The unique identifier for the DLP engine

        :return: (json)
        """
        url = "/dlpEngines"
        if dlp_engine_id:
            url = f"/dlpEngines/{dlp_engine_id}"

        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_dlp_exact_data_match_schemas(self) -> json:
        """
        Exact Data Match (EDM) templates (or EDM schemas) allow the Zscaler service to identify a record from a
        structured data source that matches predefined criteria. Using the Index Tool, you can create an EDM template
        that allows you to define the criteria (i.e., define the tokens) for your data records by importing a CSV
        file. After the data is defined and submitted within the tool, you can then apply the EDM template to a custom
        DLP dictionary or engine. This endpoint gets the EDM templates for all Index Tools used by the organization

        :return: (json)
        """
        url = "/dlpExactDataMatchSchemas"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_dlp_notification_templates(
        self,
        template_id: int = None,
    ) -> json:
        """
        Gets a list of DLP notification templates

        :param template_id: (int) Optional value. The unique identifier for a DLP notification template

        :return: (json)
        """
        url = "/dlpNotificationTemplates"
        if template_id:
            url = f"/dlpNotificationTemplates/{template_id}"

        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def add_dlp_notification_templates(
        self,
        name: str,
        subject: str,
        plain_text_message: str,
        html_message: str,
        attach_content: bool = True,
        tls_enabled: bool = True,
    ) -> json:
        """
        :param name: (str) The DLP notification template name
        :param subject: (str) The Subject line that is displayed within the DLP notification template
        :param plain_text_message: (str) The template for the plain text UTF-8 message body that must be displayed in
        the DLP notification email.
        :param html_message: (str) The template for the HTML message body that myst tbe displayed in the DLP
        notification email
        :param attach_content: (bool) if set to True, the content that is violation is attached to the DLP
        notification email
        :param tls_enabled: (bool) If set to True tls will be used to send email.

        :return: (json)
        """
        url = "/dlpNotificationTemplates"
        payload = {
            "name": name,
            "subject": subject,
            "tlsEnabled": tls_enabled,
            "attachContent": attach_content,
            "plainTextMessage": plain_text_message,
            "htmlMessage": html_message,
        }
        response = self.hp_http.post_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def delete_dlp_notification_templates(
        self,
        template_id: int,
    ) -> requests.Response:  # TODO: return json instead
        """
        Deletes a DLP notification template

        :param template_id: (int) The unique identifies for the DLP notification template
        :return: (requests.Response Object)
        """
        url = f"/dlpNotificationTemplates/{template_id}"
        response = self.hp_http.delete_call(
            url=url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response

    def list_icap_server(
        self,
        icap_server_id: int = None,
    ) -> json:
        """
        Gets a list of DLP notification templates

        :param icap_server_id: (int) Optional value. The unique identifier for the DLP server using ICAP

        :return: (json)
        """
        url = "/icapServers"
        if icap_server_id:
            url = f"/icapServers/{icap_server_id}"

        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_idm_profile(
        self,
        profile_id: int = None,
    ) -> json:
        """
        List all the IDM templates for all Index Tools used by the organization. If profileId, it lists the IDM
        template information for the specified ID.

        :param profile_id: (int) Optional value. The unique identifier for the IDM template (or profile)

        :return: (json)
        """
        if profile_id:
            url = f"/idmprofile/{profile_id}"
        else:
            url = "/idmprofile"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_web_dlp_rules(
        self,
        rule_id: int = None,
    ) -> json:
        """
        list DLP policy rules, excluding SaaS Security API DLP policy rules. If ruleId, list DLP policy rule
        information for the specified ID

        :param rule_id: (int) Optional value. The unique identifier for theDLP rule

        :return: (json)
        """
        url = "/webDlpRules"
        if rule_id:
            url = f"/webDlpRules/{rule_id}"

        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def delete_web_dlp_rules(
        self,
        rule_id: int,
    ) -> json:
        """
        Deletes a DLP policy rule. This endpoint is not applicable to SaaS Security API DLP policy rules.

        :param rule_id: (int) The unique identifier for the DLP policy rule.

        :return: (json)
        """
        url = f"/webDlpRules/{rule_id}"
        response = self.hp_http.delete_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    # Firewall Policies

    def list_network_services_lite(
        self,
    ) -> json:
        """
        Gets a summary list of all network service groups.

        :return: (json)
        """
        response = self.hp_http.get_call(
            "/networkServices/lite",
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_network_services(
        self,
        service_id: int = None,
    ) -> json:
        """
        Gets a list of all network service groups. The search parameters find matching values within the "name" or
        "description" attributes.

        :param service_id: (int)

        :return: (json)
        """
        url = "/networkServices"
        if service_id:
            url = f"/networkServices/{service_id}"

        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def add_network_services(
        self,
        name: str,
        tag: str = None,
        src_tcp_ports: list = None,
        dest_tcp_ports: list = None,
        src_udp_ports: list = None,
        dest_udp_ports: list = None,
        service_type: str = "CUSTOM",
        description: str = None,
        is_name_l10n_tag: bool = False,
    ) -> requests.Response:  # TODO: return json
        """
        Adds a new network service.

        :param name: (str) Name
        :param tag: (str)
        :param src_tcp_ports:(list) Each element is [{"start": int, "end": int}]
        :param dest_tcp_ports:(list) Each element is [{"start": int, "end": int}]
        :param src_udp_ports:(list) Each element is [{"start": int, "end": int}]
        :param dest_udp_ports:(list) Each element is [{"start": int, "end": int}]
        :param service_type: (str) STANDARD|PREDEFINE|CUSTOM
        :param description: (str) Description
        :param is_name_l10n_tag: (bool)

        :return: (requests.Response Object)
        """
        url = "/networkServices"
        payload = {
            "id": 0,
            "name": name,
            "tag": tag,
            "srcTcpPorts": src_tcp_ports,
            "destTcpPorts": dest_tcp_ports,
            "srcUdpPorts": src_udp_ports,
            "destUdpPorts": dest_udp_ports,
            "type": service_type,
            "description": description,
            "isNameL10nTag": is_name_l10n_tag,
        }
        response = self.hp_http.post_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response

    def delete_network_services(
        self,
        service_id: int,
    ) -> requests.Response:
        """
        :param service_id: (int) The unique identifier for the network service

        :return: (requests.Response Object)
        """
        url = f"/networkServices/{service_id}"
        response = self.hp_http.delete_call(
            url,
            cookies=self.cookies,
            error_handling=False,
            headers=self.headers,
        )

        return response

    def list_firewall_filtering_rules(
        self,
        rule_id: int = None,
    ) -> json:
        """
        Gets all rules in the Firewall Filtering policy.

        :param rule_id: (int)

        :return: (json)
        """
        url = "/firewallFilteringRules"
        if rule_id:
            url = f"/firewallFilteringRules/{rule_id}"

        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def add_firewall_filtering_rules(
        self,
        name: str,
        order: int,
        state: str,
        action: str,
        description: str = None,
        default_rule: bool = False,
        predefined: bool = False,
        src_ips: list = None,
        dest_addresses: list = None,
        dest_ip_groups: list = None,
        src_ip_groups: list = None,
        dest_ip_categories: list = None,
        labels=None,
        nw_services: list = None,
        rank: int = 0,
    ) -> requests.Response:
        """
        :param name: (str) Name of the Firewall Filtering policy rule ["String"]
        :param order: (int), Rule order number of the Firewall Filtering policy rule
        :param state: (str) Possible values : DISABLED or  ENABLED
        :param action: (str) Possible values: ALLOW, BLOCK_DROP, BLOCK_RESET, BLOCK_ICMP, EVAL_NWAPP
        :param description: (str) Additional information about the rule
        :param default_rule: (bool) Default is false.If set to true, the default rule is applied
        :param predefined: (bool)
        :param src_ips: (list) List of source IP addresses
        :param dest_addresses: (list) List of destination addresses
        :param dest_ip_groups: (list) List of user-defined destination IP address groups
        :param src_ip_groups: (list) List of user defined source IP address groups
        :param dest_ip_categories:(list) list of destination IP categories
        :param labels: (?)
        :param nw_services: (list) List of user-defined network services on with the rule is applied
        :param rank: (int), Admin rank of the Firewall Filtering policy rule

        :return: (requests.Response Object)
        """
        url = "/firewallFilteringRules"
        payload = {
            "accessControl": "READ_WRITE",
            "enableFullLogging": False,
            "name": name,
            "order": order,
            "rank": rank,
            "action": action,
            "state": state,
            "predefined": predefined,
            "defaultRule": default_rule,
            "description": description,
        }
        if src_ips:
            payload.update(srcIps=src_ips)
        if src_ip_groups:
            payload.update(srcIpGroups=src_ip_groups)
        if dest_addresses:
            payload.update(destAddresses=dest_addresses)
        if dest_ip_groups:
            payload.update(destIpGroups=dest_ip_groups)
        if labels:
            payload.update(labels=labels)
        if dest_ip_categories:
            payload.update(destIpCategories=dest_ip_categories)
        if nw_services:
            payload.update(nwServices=nw_services)
        response = self.hp_http.post_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response

    def delete_firewall_filtering_rules(
        self,
        rule_id: int,
    ) -> requests.Response:
        """
        Deletes a Firewall Filtering policy rule for the specified ID.

        :param rule_id: (int) The unique identifier for the policy rule

        :return: (requests.Response Object)
        """
        url = f"/firewallFilteringRules/{rule_id}"
        response = self.hp_http.delete_call(
            url,
            cookies=self.cookies,
            error_handling=False,
            headers=self.headers,
        )

        return response

    def list_ip_source_groups(
        self,
        ip_group_id: int = None,
    ) -> json:
        """
        Gets a list of all IP source groups. The search parameters find matching values within the "name" or
        "description" attributes.

        :param ip_group_id: (int) Option ip group id

        :return: (json)
        """
        url = "/ipSourceGroups"
        if ip_group_id:
            url = f"/ipSourceGroups/{ip_group_id}"

        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_ip_source_groups_lite(self) -> json:
        """
        Gets a name and ID dictionary of all IP source groups

        :return: (json)
        """
        url = "/ipSourceGroups/lite"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_ip_destination_groups(
        self,
        ip_group_id: int = None,
    ) -> json:
        """
        Gets a list of all IP source groups. The search parameters find matching values within the "name" or
        "description" attributes.

        :param ip_group_id: (int) Option ip group id

        :return: (json)
        """
        url = "/ipDestinationGroups/"
        if ip_group_id:
            url = f"/ipDestinationGroups/{ip_group_id}"

        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_ip_destination_groups_lite(self) -> json:
        """
        Gets a name and ID dictionary of all IP destination groups

        :return: (json)
        """
        url = "/ipDestinationGroups/lite"
        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def add_ip_source_groups(
        self,
        name: str,
        ip_addresses: list,
        description: str = None,
    ) -> json:
        """
        :param name: (str) Name
        :param ip_addresses: (list) List of IP addresses
        :param description: (str) description

        :return: (json)
        """
        url = "/ipSourceGroups"
        payload = {
            "name": name,
            "ipAddresses": ip_addresses,
            "description": description,
        }
        response = self.hp_http.post_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def delete_ip_source_groups(
        self,
        ip_group_id: int,
    ) -> requests.Response:
        """
        Deletes the IP source group for the specified ID

        :param ip_group_id: (int) The unique identifies for the IP source group

        :return: (requests.Response Object)
        """
        url = f"/ipSourceGroups/{ip_group_id}"
        response = self.hp_http.delete_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            payload={},
            headers=self.headers,
        )

        return response

    def delete_ip_destination_groups(
        self,
        ip_group_id: int,
    ) -> requests.Response:
        """
        Deletes the IP destination group for the specified ID

        :param ip_group_id: (int) The unique identifies for the IP source group

        :return: (requests.Response Object)
        """
        url = f"/ipDestinationGroups/{ip_group_id}"
        response = self.hp_http.delete_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            payload={},
            headers=self.headers,
        )

        return response

    def add_ip_destination_groups(
        self,
        name: str,
        dest_ip_group_type: str,
        addresses: list,
        ip_categories: list = None,
        countries: list = None,
        description: str = None,
    ) -> json:
        """
        :param name: (str) Name
        :param dest_ip_group_type: (str) Destination IP group type. Either DSTN_IP or  DSTN_FQDN or DSTN_DOMAIN
        :param addresses: (list) List of Destination IP addresses within the group.
        :param ip_categories: (list) List of Destination IP address URL categories. You can identify destination based
        on the URL category of the domain. Default value ANY
        :param countries: (list) List of destination IP address countries. You can identify destinations based on
        the location of a server.Default value ANY
        :param description: (str) description
        """
        if dest_ip_group_type not in [
            "DSTN_IP",
            "DSTN_FQDN",
            "DSTN_DOMAIN",
        ]:
            raise ValueError("Invalid destination type ")
        if countries:
            for i in countries:
                if i not in valid_countries:
                    raise ValueError("Invalid country ")
        else:
            countries = []

        if ip_categories:
            for j in ip_categories:
                if j not in valid_category_ids:
                    raise ValueError("Invalid country ")
        else:
            ip_categories = []

        url = "/ipDestinationGroups"
        payload = {
            "name": name,
            "type": dest_ip_group_type,
            "addresses": addresses,
            "ipCategories": ip_categories,
            "countries": countries,
            "description": description,
        }
        response = self.hp_http.post_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    # Device Groups

    def list_devices_groups(
        self,
        query: str = None,
    ) -> json:
        """
        Gets a list of device groups

        :param query: (str)

        :return: (json)
        """
        url = "/deviceGroups"
        if query:
            url = f"/deviceGroups?{query}"

        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def list_devices(
        self,
        query: str = None,
    ) -> json:
        """
        Gets a list of devices. Any given search parameters will be applied during device search. Search parameters
        are based on device name, model, owner, OS type, and OS version. The devices listed can also be restricted
        to return information only for ones belonging to specific users.

        :param query: (str)

        :return: (json)
        """
        url = "/deviceGroups/devices"
        if query:
            url = f"/deviceGroups/devices?{query}"

        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    # Rule Labels
    def list_rule_labels(
        self,
        rule_label_id: int = None,
    ) -> json:
        """
        Gets rule label information for the specified ID

        :param rule_label_id: (int)

        :return: (json)
        """
        url = "/ruleLabels?pageSize=1000"
        if rule_label_id:
            url = f"/ruleLabels/{rule_label_id}"

        response = self.hp_http.get_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def add_rule_label(
        self,
        name: str,
        description: str = "",
        payload: dict = None,
    ) -> json:
        """
        Adds new rule labels with the given name
        :param name: (str) name  # FIXME: Not in passed attributes.
        :param description: (str) description  # FIXME: Not in passed attributes.
        :param payload: (dict)
        """
        url = "/ruleLabels"
        if not payload:
            payload = {"name": name, "description": description}
        response = self.hp_http.post_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
            payload=payload,
        )

        return response.json()

    def delete_rule_label(self, rule_id: str):
        url = f"/ruleLabels/{rule_id}"
        response = self.hp_http.delete_call(
            url,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )
        return response

    def update_call(
        self,
        url: str,
        payload: json,
    ) -> json:
        """
        Generic PUT call. This call will overwrite all the configuration with the new payload

        :param url: (str) url of Zscaler API call
        :param payload: (json) Payload

        :return: (json)
        """
        response = self.hp_http.put_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()

    def add_call(
        self,
        url: str,
        payload: json,
    ) -> json:
        """
        Generic POST call. This call will add all the configuration with the new payload

        :param url: (str) url of Zscaler API call
        :param payload: (json) Payload

        :return: (json)
        """
        response = self.hp_http.post_call(
            url,
            payload=payload,
            cookies=self.cookies,
            error_handling=True,
            headers=self.headers,
        )

        return response.json()



import json

import requests

from zscaler_api_talkers.helpers import HttpCalls, setup_logger

logger = setup_logger(name=__name__)


class ZpaTalker(object):
    """
    ZPA API talker
    Documentation: https://help.zscaler.com/zpa/zpa-api/api-developer-reference-guide
    """

    def __init__(
        self,
        customer_id: int,
        cloud: str = "https://config.private.zscaler.com",
        client_id: str = None,
        client_secret: str = "",
    ):
        """
        :param cloud: (str) Example https://config.zpabeta.net
        :param customer_id: (int) The unique identifier of the ZPA tenant
        :param client_id: (str)
        :param client_secret: (str)
        """
        self.base_uri = cloud
        self.hp_http = HttpCalls(
            host=self.base_uri,
            verify=True,
        )
        self.jsessionid = None
        self.version = "1.3"
        self.header = None
        self.customer_id = customer_id
        if client_id and client_secret:
            self.authenticate(
                client_id=client_id,
                client_secret=client_secret,
            )

    def _obtain_all_results(
        self,
        url: str,
    ) -> list:
        """
        API response can have multiple pages. This method return the whole response in a list

        :param url: (str) url

        :return: (list)
        """
        result = []
        if "?pagesize" not in url:
            url = f"{url}?pagesize=500"  # TODO: Move to parameters
        response = self.hp_http.get_call(
            url,
            headers=self.header,
            error_handling=True,
        )
        if "list" not in response.json().keys():
            return []
        if int(response.json()["totalPages"]) > 1:
            i = 0
            while i <= int(response.json()["totalPages"]):
                result = (
                    result
                    + self.hp_http.get_call(
                        f"{url}&page={i}",
                        headers=self.header,
                        error_handling=True,
                    ).json()["list"]
                )
                i += 1
        else:
            result = response.json()["list"]

        return result

    def authenticate(
        self,
        client_id: str,
        client_secret: str,
    ) -> None:
        """
        Method to obtain the Bearer Token. Refer to https://help.zscaler.com/zpa/adding-api-keys
        :param client_id: (str) client id
        :param client_secret. (str) client secret

        return (json))
        """
        url = f"/signin"
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        payload = {
            "client_id": client_id,
            "client_secret": client_secret,
        }
        response = self.hp_http.post_call(
            url,
            headers=headers,
            error_handling=True,
            payload=payload,
            urlencoded=True,
        )
        self.header = {
            "Authorization": f"{response.json()['token_type']} {response.json()['access_token']}"
        }

        return

    # app-server-controller

    def list_servers(
        self,
        query: str = False,
        server_id: int = None,
    ) -> json:
        """
        Method to obtain all the configured Servers.

        :param query: (str) Example ?page=1&pagesize=20&search=consequat
        :param server_id: (int) Unique server id number

        :return: (json)
        """
        if server_id:
            url = (
                f"/mgmtconfig/v1/admin/customers/{self.customer_id}/server/{server_id}"
            )
        else:
            if not query:
                query = "?pagesize=500"
            url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/server{query}"
        response = self.hp_http.get_call(
            url,
            headers=self.header,
            error_handling=True,
        )

        return response.json()

    # application-controller
    def list_application_segments(
        self,
        application_id: int = None,
    ) -> json or list:
        """
        Method to obtain application segments

        :param application_id: (int) Application unique identified id

        :return: (json|list)
        """
        if application_id:
            url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/application/{application_id}"
            response = self.hp_http.get_call(
                url,
                headers=self.header,
                error_handling=True,
            )
            return response.json()

        url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/application"
        response = self._obtain_all_results(url)

        return response

    def add_application_segment(
        self,
        name: str,
        health_reporting: str,
        domain_names: list,
        segment_group_id: str,
        server_groups: list,
        common_apps_dto: list = None,
        segment_group_name: str = "",
        health_check_type: str = "DEFAULT",
        clientless_apps: list = None,
        inspection_apps: list = None,
        sra_apps: list = None,
        tcp_port_range: dict = None,
        tcp_port_ranges: list = None,
        udp_port_ranges: list = None,
        udp_port_range: dict = None,
        description: str = "",
        enabled: bool = True,
        icmp_access_type: str = "NONE",
        ip_anchored: bool = False,
        double_encrypt: bool = False,
        bypass_type: str = "NEVER",
        is_cname_enabled: bool = True,
        select_connector_close_to_app: bool = False,
        passive_health_enabled: bool = True,
    ) -> json:
        """
        Adds a new Application Segment for a ZPA tenant.
        :param name: (str) App Name
        :param health_reporting: (str) possible values: NONE, ON_ACCESS, CONTINUOUS
        :param domain_names: (list) List of domains or IP addresses
        :param segment_group_id: (str) Application Segment Group id
        :param server_groups=(list) List of dictionaries, where key is id and value is serverGroupId [
        {"id": "<serverGroupId>"}
        ]
        :param common_apps_dto: (list) List of dictionaries, where appsConfig will list the apps with Browser Access
        or Inspection
        :param segment_group_name: (str) Application Segment Group Name
        :param health_check_type: (str)
        :param clientless_apps: (list) List of application domains in Application Segment with Browser access enabled
        :param inspection_apps: (list) List of application domains in Application Segment with Inspection enabled
        :param sra_apps: (list) List of application domains in Application Segment with Privileged Remote Access enabled
        :param tcp_port_range: type dict.  [{"from":int, "to":int}]
        :param tcp_port_ranges: (list)  ["from", "to"]. This will be deprecated in the future.
        :param udp_port_range: type dict.  [{"from":int, "to":int}]
        :param udp_port_ranges: (list)  ["from", "to"]. This will be deprecated in the future.
        :param description: (str) Description
        :param enabled: (bool) (True|False)
        :param icmp_access_type: (str) possible values: PING_TRACEROUTING, PING, NONE
        :param ip_anchored: (bool) (True|False)
        :param double_encrypt: (bool) (True|False)
        :param bypass_type: (str) possible values ALWAYS, NEVER, ON_NET
        :param is_cname_enabled: (bool) (True|False)
        :param select_connector_close_to_app: (bool) (True|False)
        :param passive_health_enabled: (bool) (True|False)

        :return: (json)
        """

        url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/application"
        payload = {
            "name": name,
            "description": description,
            "enabled": enabled,
            "healthCheckType": health_check_type,
            "healthReporting": health_reporting,
            "icmpAccessType": icmp_access_type,
            "ipAnchored": ip_anchored,
            "doubleEncrypt": double_encrypt,
            "bypassType": bypass_type,
            "isCnameEnabled": is_cname_enabled,
            "clientlessApps": clientless_apps,
            "inspectionApps": inspection_apps,
            "sraApps": sra_apps,
            "commonAppsDto": common_apps_dto,
            "selectConnectorCloseToApp": select_connector_close_to_app,
            "passiveHealthEnabled": passive_health_enabled,
            "tcpPortRanges": tcp_port_ranges,
            "tcpPortRange": tcp_port_range,
            "udpPortRange": udp_port_range,
            "udpPortRanges": udp_port_ranges,
            "domainNames": domain_names,
            "segmentGroupId": segment_group_id,
            "segmentGroupName": segment_group_name,
            "serverGroups": server_groups,
        }
        response = self.hp_http.post_call(
            url=url,
            payload=payload,
            headers=self.header,
            error_handling=True,
        )

        return response.json()

    def update_application_segment(
        self,
        application_id: int,
        payload: dict,
    ) -> requests.Response:
        """
        Updates the Application Segment details for the specified ID

        :param application_id: (int) Application ID
        :param payload: (dict)

        :return: (requests.Response Object)
        """
        url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/application/{application_id}"
        response = self.hp_http.put_call(
            url=url,
            payload=payload,
            headers=self.header,
            error_handling=True,
        )

        return response

    def delete_application_segment(
        self,
        application_id: int,
    ) -> requests.Response:
        """
        Updates the Application Segment details for the specified ID

        :param application_id: (int) Application ID

        :return: (requests.Response Object)
        """
        url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/application/{application_id}"
        response = self.hp_http.delete_call(
            url=url,
            error_handling=True,
        )

        return response

    # segment-group-controller

    def list_segment_group(
        self,
        segment_group_id: int = None,
        query: str = False,
    ) -> json or list:
        """
        Get all the configured Segment Groups. If segmentGroupId obtains the segment sroup details

        :param segment_group_id: (int) The unique identifier of the Segment Group.
        :param query: (str) Example ?page=1&pagesize=20&search=consequat

        return (json|list)
        """
        if segment_group_id:
            url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/segmentGroup/{segment_group_id}"
            response = self.hp_http.get_call(
                url, headers=self.header, error_handling=True
            ).json()
        else:
            if not query:
                query = "?pagesize=500"
            url = (
                f"/mgmtconfig/v1/admin/customers/{self.customer_id}/segmentGroup{query}"
            )
            response = self._obtain_all_results(url)

        return response

    def add_segment_group(
        self,
        name: str,
        description: str,
        enabled: bool = True,
    ) -> json:
        """
        Add a new segment group

        :param name: (str) Name of segment Group
        :param description: (str) Description
        :param enabled: (bool): True or False
        :return: (json)
        """
        url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/segmentGroup"
        payload = {
            "name": name,
            "description": description,
            "enabled": enabled,
        }
        response = self.hp_http.post_call(
            url,
            headers=self.header,
            error_handling=True,
            payload=payload,
        )

        return response.json()

    def delete_segment_group(self, segmentGroupId: int) -> json:
        """
        Deletes specified Segment Group.
        :param segmentGroupId: The unique identifier of the Segment Group.
        return: response
        """
        url: str = f'/mgmtconfig/v1/admin/customers/{self.customerId}/segmentGroup/{segmentGroupId}'
        response = self.hp_http.delete_call(url=url, error_handling=True)
        return response

    def update_segment_group(self, segmentGroupId: int, payload: object) -> json:
        """
        Update Segment Group
        :param segmentGroupId: type int. The unique identifier of the Segment Group.
        :param payload: type dict. Segment Group details to be updated.
        :return: Json
        """
        url: str = f'/mgmtconfig/v1/admin/customers/{self.customerId}/segmentGroup/{segmentGroupId}'
        response = self.hp_http.put_call(url, headers=self.header, error_handling=True, payload=payload)
        return response

    # connector-controller
    def list_connector(
        self,
        connector_id: int = None,
    ) -> json or list:
        """
        Get all the configured Segment Groups. If segmentGroupId obtains the segment group details

        :param connector_id: The unique identifier of the App Connector.

        return (json|list)
        """
        if connector_id:
            url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/connector/{connector_id}"
            return self.hp_http.get_call(
                url,
                headers=self.header,
                error_handling=True,
            ).json()
        else:
            url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/connector"
            response = self._obtain_all_results(url)

        return response

    def delete_bulk_connector(
        self,
        ids: list,
    ) -> json:
        """
        Get all the configured Segment Groups. If segmentGroupId obtains the segment sroup details

        :param ids: (list) list of resources ids for bulk deleting the App Connectors.

        return (json)
        """
        url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/connector/bulkDelete"
        payload = {"ids": ids}
        response = self.hp_http.post_call(
            url=url,
            headers=self.header,
            error_handling=True,
            payload=payload,
        )

        return response.json()

    # Connector-group-controller
    def list_connector_group(
        self,
        app_connector_group_id: int = None,
    ) -> json or list:
        """
        Gets all configured App Connector Groups for a ZPA tenant.

        :param app_connector_group_id: (int) The unique identifier of the Connector Group.

        return (json|list)
        """
        if app_connector_group_id:
            url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/appConnectorGroup/{app_connector_group_id}"
            return self.hp_http.get_call(
                url,
                headers=self.header,
                error_handling=True,
            ).json()
        else:
            url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/appConnectorGroup"
            response = self._obtain_all_results(url)

        return response

    def add_connector_group(self, name: str, description: str, latitude: str, longitude: str, location: str, upgradeDay: str = 'SUNDAY',
                            enabled: bool = True,
                            dnsQueryType: str = 'IPV4_IPV6', upgradeTimeInSecs: int = 66600,
                            overrideVersionProfile: bool = False, versionProfileId: int = None, tcpQuickAckApp: bool = False,
                            tcpQuickAckAssistant: bool = False, tcpQuickAckReadAssistant: bool = False, cityCountry: str = "",
                            countryCode: str = "", connectors: list = [], serverGroups: list = [], lssAppConnectorGroup: bool = False) -> json:
        """
        :param name: type string. Name of App Connector Group
        :param description: type string. Description
        :param latitude: type string. Latitude of App Connector Group
        :param longitude: type string. Longitude of App Connector Group
        :param location: type string. Location of the App Connector Group
        :param upgradeDay: type string. App Connectors in this group attempt to update to a newer version of the software during this specified day
        :param upgradeTimeInSecs: type int. App Connectors in this group attempt to update to a newer version of the software during this specified time
        :param overrideVersionProfile: type boolean. Whether the default version profile of the App Connector Group is applied or overridden
        :param versionProfileId: type int. ID of the version profile
        :param tcpQuickAckApp: type boolean. Whether TCP Quick Acknowledgement is enabled or disabled for the application. The tcpQuickAckApp, tcpQuickAckAssistant, and tcpQuickAckReadAssistant fields must all share the same value.
        :param tcpQuickAckAssistant: type boolean. Whether TCP Quick Acknowledgement is enabled or disabled for the application
        :param tcpQuickAckReadAssistant: type boolean. Whether TCP Quick Acknowledgement is enabled or disabled for the application
        :param connectors: type dict. App Connector Id's part of the App Connector Group.
        :param serverGroups: type dict. Server Groups part of App Connector Group
        :param lssAppConnectorGroup: type boolean. Is App Connector Group reserved for LSS
        """
        url: str = f'/mgmtconfig/v1/admin/customers/{self.customerId}/appConnectorGroup'
        payload: dict[str | Any, object | Any] = {
            "name": name,
            "description": description,
            "latitude": latitude,
            "longitude": longitude,
            "location": location,
            "upgradeDay": upgradeDay,
            "enabled": enabled,
            "dnsQueryType": dnsQueryType,
            "upgradeTimeInSecs": upgradeTimeInSecs,
            "overrideVersionProfile": overrideVersionProfile,
            "versionProfileId": versionProfileId,
            "tcpQuickAckApp": tcpQuickAckApp,
            "tcpQuickAckAssistant": tcpQuickAckAssistant,
            "tcpQuickAckReadAssistant": tcpQuickAckReadAssistant,
            "cityCountry": cityCountry,
            "countryCode": countryCode,
            "connectors": connectors,
            "serverGroups": serverGroups,
            "lssAppConnectorGroup": lssAppConnectorGroup
        }
        response = self.hp_http.post_call(url, headers=self.header, error_handling=True, payload=payload)
        return response.json()

    def update_connector_group(self, appConnectorGroupId: int, payload: dict) -> json:
        """
        Update configured App Connector Groups for a ZPA tenant.
        :param appConnectorGroupId: type int. The unique identifier of the Connector Group
        :param payload: type dict. Details of App Connector group to be updated
        return response
        """
        url: str = f'/mgmtconfig/v1/admin/customers/{self.customerId}/appConnectorGroup/{appConnectorGroupId}'
        response = self.hp_http.put_call(url, headers=self.header, error_handling=True, payload=payload)
        return response

    def delete_connector_group(self, appConnectorGroupId: int) -> json:
        """
            Delete specified App Connector Group
            :param appConnectorGroupId: type int. The unique identifier of the Connector Group
            return response
        """
        url: str = f'/mgmtconfig/v1/admin/customers/{self.customerId}/appConnectorGroup/{appConnectorGroupId}'
        response = self.hp_http.delete_call(url, error_handling=True)
        return response
    # ba-certificate-controller-v-2

    def list_browser_access_certificates(
        self,
    ) -> list:  # FIXME: duplicate but URL is slightly different.
        """
        Get all Browser issued certificates

        :return: (list)
        """
        url = f"/mgmtconfig/v2/admin/customers/{self.customer_id}/clientlessCertificate/issued"
        response = self._obtain_all_results(url)

        return response

    # enrollment-cert-controller

    def list_enrollment_certificates(self) -> list:
        """
        Get all the Enrollment certificates

        :return: (list)
        """
        url = f"/mgmtconfig/v2/admin/customers/{self.customer_id}/enrollmentCert"
        response = self._obtain_all_results(url)

        return response

    def list_v1_browser_access_certificates(
        self,
    ) -> list:
        """
        Get all the issued certificates

        :return: (list)
        """
        url = (
            f"/mgmtconfig/v1/admin/customers/{self.customer_id}/visible/versionProfiles"
        )
        response = self._obtain_all_results(url)

        return response

    # customer-version-profile-controller

    def list_customer_version_profile(
        self,
        query: str = False,
    ) -> json:
        """
        Get Version Profiles visible to a customer

        :param query: (str) Example ?page=1&pagesize=20&search=consequat

        :return: (json)
        """
        if not query:
            query = "?pagesize=500"
        url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/visible/versionProfiles{query}"
        response = self.hp_http.get_call(
            url,
            headers=self.header,
            error_handling=True,
        )

        return response.json()

    # cloud - connector - group - controller
    def list_cloud_connector_group(
        self,
        group_id: int = None,
        query: str = False,
    ) -> json:
        """
        Get all configured Cloud Connector Groups. If id, Get the Cloud Connector Group details

        :param group_id: (int)
        :param query: (str) Example ?page=1&pagesize=20&search=consequat

        :return: (json)
        """
        if group_id:
            url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/cloudConnectorGroup/{group_id}"
        else:
            if not query:
                query = "?pagesize=500"
            url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/cloudConnectorGroup{query}"

        response = self.hp_http.get_call(
            url,
            headers=self.header,
            error_handling=True,
        )

        return response.json()

    # idp-controller-v-2
    def list_idp(
        self,
        query: str = False,
    ) -> list:
        """
        Method to Get all the idP details for a ZPA tenant

        :param query: (str) HTTP query

        :return: (list)
        """
        if not query:
            query = "?pagesize=500"
        url = f"/mgmtconfig/v2/admin/customers/{self.customer_id}/idp{query}"
        response = self._obtain_all_results(url)

        return response

    # provisioningKey-controller
    def list_provisioning_key(
        self,
        association_type: str = "CONNECTOR_GRP",
    ) -> list:
        """
        Gets details of all the configured provisioning keys.

        :param association_type: (str) The supported values are CONNECTOR_GRP and SERVICE_EDGE_GRP.

        :return: (list)
        """
        url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/associationType/{association_type}/provisioningKey"
        response = self._obtain_all_results(url)

        return response

    # policy-set-controller

    # scim-attribute-header-controller

    def list_scim_attributes(
        self,
        idp_id: int,
        query: str = False,
    ) -> json:
        """
        :param idp_id: (int) The unique identifies of the Idp
        :param query: (str) ?page=1&pagesize=20&search=consequat

        :return: (json)
        """
        if not query:
            query = "?pagesize=500"
        url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/idp/{idp_id}/scimattribute{query}"
        response = self.hp_http.get_call(
            url,
            headers=self.header,
            error_handling=True,
        )

        return response.json()

    # scim-group-controller
    def list_scim_groups(
        self,
        idp_id: int,
        query: str = False,
    ) -> list:
        """
        Method to list all SCIM groups

        :param idp_id: (int) The unique identifies of the Idp
        :param query: (str) ?page=1&pagesize=20&search=consequat

        :return: (list)
        """
        if not query:
            query = "?pagesize=500"
        url = f"/userconfig/v1/customers/{self.customer_id}/scimgroup/idpId/{idp_id}{query}"
        response = self._obtain_all_results(url)

        return response

    # saml-attr-controller-v-2
    def list_saml_attributes(self) -> list:
        """
        Method to get all SAML attributes

        :return: (list)
        """
        url = f"/mgmtconfig/v2/admin/customers/{self.customer_id}/samlAttribute"
        response = self._obtain_all_results(url)

        return response

    # global-policy-controller

    def list_policies(
        self,
        policy_type: str = "ACCESS_POLICY",
    ) -> list:
        """list policie(s)  by policy type,

        :param policy_type: (str) Supported values Possible values = ACCESS_POLICY,GLOBAL_POLICY, TIMEOUT_POLICY,
        REAUTH_POLICY, SIEM_POLICY, CLIENT_FORWARDING_POLICY,BYPASS_POLICY

        :return: (list)
        """
        url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/policySet/rules/policyType/{policy_type}"
        response = self._obtain_all_results(url)

        return response

    def list_policy_set(
        self,
        policy_type: str = "ACCESS_POLICY",
    ) -> json:
        """Gets the policy set for the specified policy type

        :param policy_type: (str) Supported values are ACCESS_POLICY,GLOBAL_POLICY, TIMEOUT_POLICY,REAUTH_POLICY,
        SIEM_POLICY, CLIENT_FORWARDING_POLICY,BYPASS_POLICY

        :return: (json)
        """
        url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/policySet/policyType/{policy_type}"
        response = self.hp_http.get_call(
            url,
            headers=self.header,
            error_handling=True,
        )

        return response.json()

    def add_policy_set(
        self,
        app_operands: list,
        rule_name: str,
        action: str,
        policy_set_id: int,
        operands: list,
        operator: str,
        msg_string: str = None,
    ) -> json:
        """
        Method to create a new access Policy

        :param app_operands: (list) List of app_operands: Examples = [{
        "objectType": "APP",
        "lhs": "id",
        "rhs": applicationId,
        }]
        :param rule_name: (str) Policy set Rule Name
        :param action: (str) ALLOW / DENY
        :param policy_set_id: (int) Global Policy ID. can be obtained from list_global_policy_id
        :param operands: (list) List of operands. Example = [{
        "objectType": "SAML",
        "lhs": "<samlAttrId>",
        "rhs": "<samlAttrValue>",
        },{
        "objectType": "SCIM",
        "lhs": "<scimAttrId>",
        "rhs": "<scimAttrValue>”
        }]
        :param operator: (str)
        :param msg_string: (str)

        :return: (json)
        """
        url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/policySet/{policy_set_id}/rule"
        payload = {
            "conditions": [
                {"operands": app_operands},
                {
                    "operands": operands,
                    "operator": operator,
                },
            ],
            # Seems here needs to be AND
            "operator": "AND",
            "name": rule_name,
            "description": "Description",
            "action": action,
            "customMsg": msg_string,
        }
        logger.info(payload)
        response = self.hp_http.post_call(
            url=url,
            headers=self.header,
            error_handling=True,
            payload=payload,
        )

        return response.json()

    # Server Group Controller

    def list_server_groups(
        self,
        group_id: int = None,
    ) -> json or list:
        """
        Method to get all configured Server Groups. If groupI, get the Server Group details

        :param group_id: (int) The unique identifier of the Server Group.

        :return: (json|list)
        """
        if group_id:
            url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/serverGroup/{group_id}"
            response = self.hp_http.get_call(
                url,
                headers=self.header,
                error_handling=True,
            ).json()
        else:
            url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/serverGroup"
            response = self._obtain_all_results(url)

        return response

    def add_server_groups(
        self,
        name: str,
        description: str,
        connector_group_id: list,
    ) -> json:
        """
        :param name: (str) Server Group Name
        :param description: (str) Description
        :param connector_group_id: (list) List of dictionaries with key as "id" and value connector_group_id.
        [{"id": connector_group_id}]

        :return: (json)
        """
        url = f"/mgmtconfig/v1/admin/customers/{self.customer_id}/serverGroup"
        payload = {
            "enabled": True,
            "dynamicDiscovery": True,
            "name": name,
            "description": description,
            "servers": [],
            "appConnectorGroups": connector_group_id,
        }
        response = self.hp_http.post_call(
            url=url,
            headers=self.header,
            error_handling=True,
            payload=payload,
        )

        return response.json()

    def list_posture_profiles(
        self,
        query: str = False,
    ) -> list:
        """
        Method to Get all the idP details for a ZPA tenant

        :param query: (str) HTTP query

        :return: (list)
        """
        if not query:
            query = "?pagesize=500"
        url = f"/mgmtconfig/v2/admin/customers/{self.customer_id}/posture{query}"
        response = self._obtain_all_results(url)

        return response

    def list_privileged_consoles(
        self,
        query: str = False,
    ) -> list:
        """
        Method to Get all the privileged_remote_consoles for a ZPA tenant

        :param query: (str) HTTP query

        :return: (list)
        """
        if not query:
            query = "?pagesize=500"
        url = f"/mgmtconfig/v2/admin/customers/{self.customer_id}/privilegedConsoles{query}"
        response = self._obtain_all_results(url)

        return response

    def list_sra_consoles(self) -> list:
        """
        Method to obtain list of sra consoles from all application segments

        :return: (list)
        """
        sra_list = []
        app_segments = self.list_application_segments()
        for apps in app_segments:
            srap = apps.get("sraApps")
            if srap is not None:
                sra_list.extend(srap)

        return sra_list

    # Certificate Controller v2
    def list_issued_certificates(
        self,
        query: str = None,
    ) -> list:
        """
        Method to get all issued certificates

        :return: (list)
        """
        if not query:
            query = "?pagesize=500"  # TODO: Query never put into url.

        url = f"/mgmtconfig/v2/admin/customers/{self.customer_id}/certificate/issued"
        response = self._obtain_all_results(url)

        return response



import logging
import os
from logging.handlers import RotatingFileHandler


def setup_logger(
    name: str,
    level: str = os.getenv("LOGGING_LEVEL", "info"),
    log_filename: str = None,
    max_log_size: int = 10 * 1024 * 1024,  # 10 MB
    backup_count: int = 3,
) -> logging.Logger:
    """
    To set up as many loggers as you want

    Args:
        name (str): Name of logger
        level (str): os.getenv("LOGGING_LEVEL"). Defaults to 'info'.
        log_filename (str): Name and Path of log file.
        max_log_size (int): Defaults to 10MB
        backup_count (int): Default to 3

    Return:
        Logger Object
    """
    logger = logging.getLogger(name)

    formatter = logging.Formatter(
        fmt="{asctime} {name}.{funcName} {levelname} {message}",
        datefmt="%Y%m%d %H:%M:%S",
        style="{",
    )

    if log_filename:
        handler_file = RotatingFileHandler(
            log_filename,
            maxBytes=max_log_size,
            backupCount=backup_count,
        )
        handler_file.setFormatter(formatter)
        logger.addHandler(handler_file)

    handler_stream = logging.StreamHandler()
    handler_stream.setFormatter(formatter)
    logger.addHandler(handler_stream)

    log_level = logging.INFO
    level = level.lower()
    if level == "notset":
        log_level = logging.NOTSET
    elif level == "debug":
        log_level = logging.DEBUG
    elif level == "info":
        log_level = logging.INFO
    elif level == "warning" or level == "warn":
        log_level = logging.WARNING
    elif level == "error":
        log_level = logging.ERROR
    elif level == "critical":
        log_level = logging.CRITICAL
    logger.setLevel(log_level)

    return logger



import json
import pdb
import time
import requests
from zscaler_api_talkers.helpers.http_calls import HttpCalls
from zscaler_api_talkers.helpers.logger import setup_logger

logger = setup_logger(name=__name__)


class ClientConnectorTalker(object):
    """
    Client Connector API talker
    Documentation: under development
    Currently in beta status
    """

    def __init__(
            self,
            cloud: str,
            client_id: str = "",
            secret_key: str = "",
    ):
        """
        :param cloud: (str) Top Level Domain (TLD) of the Zscaler cloud where tenant resides.
        :param client_id: (str) Client ID
        :param secret_key: (str) Secret Key
        """
        self.base_uri = f"https://api-mobile.{cloud}/papi"
        self.hp_http = HttpCalls(
            host=self.base_uri,
            verify=True,
        )
        self.jsession_id = None
        self.version = "beta 0.1"
        self.header = {}
        if client_id and secret_key:
            self.authenticate(
                client_id=client_id,
                secret_key=secret_key,
            )

    def authenticate(
            self,
            client_id: str,
            secret_key: str,
    ):
        """
        Method to authenticate.

        :param client_id: (str) Client id
        :param secret_key: (str) Client secret, obtained from portal.
        """
        payload = {
            "apiKey": client_id,
            "secretKey": secret_key,
        }
        url = "/auth/v1/login"
        response = self.hp_http.post_call(
            url=url,
            headers={"Accept": "*/*"},
            payload=payload,
        )
        self.header = {"auth-token": response.json()["jwtToken"]}

    def _obtain_all(
            self,
            url: str,
            cookies: dict = None,
            params: dict = None,
            headers: dict = None,
    ) -> json:
        """
        Internal method that queries all pages

        :param url: (str) URL
        :param cookies: (dict?) Cookies
        :param params: (dict) Parameters to pass in request
        :param headers: (dict) Headers to pass in request

        :return: (json) JSON of results
        """
        page = 1
        result = []
        while True:
            response = self.hp_http.get_call(
                f"{url}&page={page}",
                cookies=cookies,
                params=params,
                headers=headers,
                error_handling=True,
            )
            if response.json():
                result += response.json()
                page += 1
                time.sleep(0.5)
            else:
                break

        return result

    def list_devices(
            self,
            username: str = None,
            os_type: str = None,
    ) -> json:
        """
        Gets the list of all enrolled devices of your organization and their basic details.
        :param username: (str) Username in email format
        :param os_type: (str)  1 - iOS, 2 - Android, 3 - Windows, 4 - macOS, 5 - Linux
        :return: (json) JSON of results
        """
        url = "/public/v1/getDevices?pageSize=500"
        if username:
            url += f'&username={username}'
        if os_type:
            url += f'&osType={os_type}'

        response = self._obtain_all(
            url=url,
            headers=self.header,
        )

        return response

    def list_otp(
            self,
            ud_id: int,
    ) -> json:
        """
        Method to fetch the One Time Password for a specific device. These passwords are unique and tied to a device
        UDID.
        :param ud_id: (int) User device ID
        :return: (json) JSON of results
        """
        url = f"/public/v1/getOtp"
        parameters = {
            "udid": ud_id,
        }
        response = self.hp_http.get_call(
            url=url,
            params=parameters,
            headers=self.header,
        )

        return response.json()

    def list_passwords(
            self,
            ud_id: int,
    ):
        """
        Method to fetch the One Time Password for a specific device. These passwords are unique and tied to a device UDID

        :param ud_id: (int) User device ID
        :return: (json) JSON of results
        """
        url = f"/public/v1/getOtp"
        parameters = {
            "udid": ud_id,
        }
        response = self.hp_http.get_call(
            url=url,
            params=parameters,
            headers=self.header,
        )

        return response.json()

    def remove_devices(
            self,
            username: str = None,
            client_connector_version: str = None,
            ud_ids: list = None,
            os_type: int = 0,
    ) -> json:
        """
        Method to  mark the device for removal (Device Removal Pending).
        API currently can remove up to 30 devices per call
        :param username: type str. Userna,e
        :param ud_ids: type list. List of user devices ids
        :param os_type: 0 ALL OS types, 1 IOS, 2 Android, 3 Windows, 4 macOS, 5 Linux
        :param client_connector_version: Client connector version
        :return: (json) JSON of results
        """
        url = f"/public/v1/removeDevices"
        payload = {
            "userName": username,
            "clientConnectorVersion": client_connector_version,
            "udids": ud_ids,
            "osType": os_type,
        }
        response = self.hp_http.post_call(
            url=url,
            headers=self.header,
            payload=payload,
        )

        return response.json()

    def force_remove_devices(
            self,
            username: str = None,
            client_connector_version: str = None,
            ud_ids: list = None,
            os_type: int = 0,
    ) -> json:
        """
        Force Remove, has the same effect as Remove, though it additionally moves the device straight to Removed and also
        signals the cloud to invalidate the user’s session.
        API currently can remove up to 30 devices per call

        :param client_connector_version: (str) ZCC version
        :param ud_ids: (list) List of user devices ids
        :param os_type: (int) 0 ALL OS types, 1 IOS, 2 Android, 3 Windows, 4 macOS, 5 Linux
        :param username:(str) Username

        :return: (json) JSON of results
        """
        if ud_ids is None:
            ud_ids = []
        url = f"/public/v1/forceRemoveDevices"
        payload = {
            "clientConnectorVersion": client_connector_version,
            "udids": ud_ids,
            "osType": os_type,
            "userName": username
        }
        response = self.hp_http.post_call(
            url=url,
            headers=self.header,
            payload=payload,
        )

        return response.json()

    def list_download_service_status(
            self,
    ) -> requests.Response.content:
        """
        Method to download Service Status
        :return: (str) String of results
        """
        url = "/public/v1/downloadServiceStatus"

        response = self.hp_http.get_call(
            url=url,
            headers=self.header,
        )

        return response.content


class ZccTalker(ClientConnectorTalker):
    def __init__(
            self,
            cloud: str,
            client_id: str = "",
            secret_key: str = "",
    ):
        logger.warning(
            "Deprecating ZccTalker. Start using ClientConnectorTalker instead."
        )
        super().__init__(
            cloud,
            client_id,
            secret_key,
        )



import requests

from .logger import setup_logger

logger = setup_logger(name=__name__)


def _zia_http_codes(response: requests.Response):
    """
    Internal method to display HTTP error handling and response codes. For more information, please refer to
    https://help.zscaler.com/zia/about-error-handling

    :param response: (requests.Response Object)
    """
    if response.status_code in [200, 201, 202, 204]:
        return
    elif response.status_code == 401:
        raise ValueError(
            f"{response.status_code} :Session is not authenticated or timed out"
        )
    elif response.status_code == 403:
        raise ValueError(
            f"{response.status_code} :API key disabled or SKU subscription missing or user role has not access"
        )
    elif response.status_code == 404:
        raise ValueError("Resource does not exist")
    elif response.status_code == 409:
        raise ValueError(
            "Request could not be processed because of possible edit conflict occurred"
        )
    elif response.status_code == 415:
        raise ValueError("Unsupported media type")
    elif response.status_code == 429:
        raise ValueError("Exceeded the rate limit or quota")
    elif response.status_code == 500:
        raise ValueError("Unexpected error")
    elif response.status_code == 503:
        raise ValueError("Service is temporarily unavailable")
    else:
        print(response.content)
        raise ValueError(f"Unexpected HTTP response code: {response.status_code}")


class HttpCalls(object):
    """
    class to Perform HTTP calls
    """

    def __init__(
        self,
        host: str,
        header: dict = None,
        verify: bool = True,
    ):
        """
        to start this instance, host IP address or fqdn is required

        :param host: (str) IP address or fqdn
        :param header: (dict) HTTP header
        :param verify: (bool) True to verify ssl cert with in HTTP call
        """
        self.version = "1.1"
        self.host = host
        self.headers = {"Content-type": "application/json", "Cache-Control": "no-cache"}
        if header:
            self.headers.update(header)
        self.cookies = None
        self.verify = verify

    def get_call(
        self,
        url: str,
        cookies: dict = None,
        headers: dict = None,
        params: dict = None,
        error_handling: bool = False,
    ) -> requests.Response:
        """
        Method to perform a GET HTTP  call

        :param url: (str) url
        :param cookies: (str) cookies
        :param headers: (dict) Additional HTTP headers
        :param params: (dict) Key,Value parameters in the URL after a question mark
        :param error_handling: (bool) when TRUE will use Zscaler HTTP codes

        :return: (requests.Response Object)
        """
        full_url = f"{self.host}{url}"
        if headers:
            self.headers.update(headers)
        try:
            response = requests.get(
                url=full_url,
                headers=self.headers,
                cookies=cookies,
                params=params,
                verify=self.verify,
            )
            if error_handling:
                _zia_http_codes(response)
            else:
                if response.status_code not in [200, 201, 204]:
                    raise ValueError(f"{response.status_code} -> {response.content}")
            return response
        except requests.HTTPError as e:
            raise ValueError(e)

    def post_call(
        self,
        url: str,
        payload: dict,
        params: dict = None,
        headers: dict = None,
        cookies: dict = None,
        error_handling: bool = False,
        urlencoded: bool = False,
    ) -> requests.Response:
        """
        Method to perform an HTTP POST call

        :param url: (str) url
        :param payload: (dict)
        :param params: (dict)
        :param headers: (dict)
        :param cookies: (str) cookies
        :param error_handling: (bool) when TRUE will use Zscaler HTTP codes
        :param urlencoded: (bool)

        :return: (requests.Response Object)
        """
        full_url = f"{self.host}{url}"
        try:
            if urlencoded:
                url_encoded_headers = headers
                response = requests.post(
                    url=full_url,
                    params=params,
                    headers=url_encoded_headers,
                    cookies=cookies,
                    data=payload,
                    verify=self.verify,
                )
            else:
                if headers:
                    self.headers.update(headers)
                response = requests.post(
                    url=full_url,
                    params=params,
                    headers=self.headers,
                    cookies=cookies,
                    json=payload,
                    verify=self.verify,
                )
            if error_handling:
                _zia_http_codes(response)
            else:
                if response.status_code not in [200, 201, 204]:
                    raise ValueError(f"{response.status_code} -> {response.content}")
            return response
        except requests.HTTPError as e:
            raise ValueError(e)

    def patch_call(
        self,
        url: str,
        payload: dict,
        cookies: dict = None,
    ) -> requests.Response:
        """
        Method to perform an HTTP PATH call

        :param url: (str) url
        :param payload: (dict)
        :param cookies: (str) cookies

        :return: (requests.Response Object)
        """
        full_url = f"{self.host}{url}"
        try:
            response = requests.patch(
                url=full_url,
                headers=self.headers,
                cookies=cookies,
                json=payload,
                verify=self.verify,
            )
            if response.status_code not in [200, 201, 204]:
                raise ValueError(response.status_code)
            return response
        except requests.HTTPError as e:
            raise ValueError(e)

    def put_call(
        self,
        url: str,
        payload: dict,
        params: dict = None,
        headers: dict = None,
        cookies: dict = None,
        error_handling: bool = False,
    ) -> requests.Response:
        """
        Method to perform an HTTP PUT call

        :param url: (str) url
        :param params: (dict) Parameters to add to url
        :param payload: (dict)
        :param headers: (dict)
        :param cookies: (str) cookies
        :param error_handling: (bool) when TRUE will use Zscaler HTTP codes

        :return: (requests.Response Object)
        """
        full_url = f"{self.host}{url}"
        if headers:
            self.headers.update(headers)
        try:
            response = requests.put(
                url=full_url,
                params=params,
                headers=self.headers,
                cookies=cookies,
                json=payload,
                verify=self.verify,
            )
            if error_handling:
                _zia_http_codes(response)
            else:
                if response.status_code not in [200, 201, 204]:
                    try:
                        raise ValueError(
                            f"HTTPS Response code {response.status_code} : {response.json()}"
                        )
                    except ValueError:
                        raise ValueError(response.status_code)
            return response
        except requests.HTTPError as e:
            raise ValueError(e)

    def delete_call(
        self,
        url: str,
        payload: dict = None,
        headers: dict = None,
        cookies: dict = None,
        error_handling: bool = False,
    ) -> requests.Response:
        """
        Method to perform an HTTP DELETE call

        :param url: (str) url
        :param payload: (dict) json payload
        :param headers: (dict)
        :param cookies: (str) cookies
        :param error_handling: (bool) when TRUE will use Zscaler HTTP codes

        :return: (requests.Response Object)
        """
        full_url = f"{self.host}{url}"
        if headers:
            self.headers.update(headers)
        try:
            response = requests.delete(
                url=full_url,
                headers=self.headers,
                cookies=cookies,
                json=payload,
                verify=self.verify,
            )
            if error_handling:
                _zia_http_codes(response)
            else:
                if response.status_code not in [200, 201, 204]:
                    raise ValueError(response.status_code)
            return response
        except requests.HTTPError as e:
            raise ValueError(e)



import time

from zscaler_api_talkers.helpers import request_


def _obfuscate_api_key(
    seed: str,
) -> (time, str):
    """
    Internal method to Obfuscate the API key

    :param seed: (str) API key

    :return: (str, str) timestamp,obfuscated key
    """
    now = int(time.time() * 1000)
    n = str(now)[-6:]
    r = str(int(n) >> 1).zfill(6)
    key = ""
    for i in range(0, len(str(n)), 1):
        key += seed[int(str(n)[i])]
    for j in range(0, len(str(r)), 1):
        key += seed[int(str(r)[j]) + 2]

    return now, key


def _get_seed(
    url: str,
) -> str:
    result = request_(
        method="get",
        url=url,
    )
    api = result.text.split('"')
    js = None
    for each in api:
        if each.startswith("js"):
            if each.find("lean") > 0:
                js = each
                break

    result = request_(
        method="get",
        url=f"{url}/{js}",
    )
    key = result.text.split(".")
    seed = None
    for each in key:
        if each.startswith("obfuscateApiKey"):
            getkey = each.split('"')
            seed = getkey[1]
            break

    return seed



from .http_calls import HttpCalls
from .logger import setup_logger
from .utilities import get_user_agent, request_

__all__ = [
    "setup_logger",
    "request_",
    "get_user_agent",
    "HttpCalls",
]



from .client_connector.talker import ClientConnectorTalker, ZccTalker
from .zia.talker import ZiaTalker
from .zpa.talker import ZpaTalker

__all__ = [
    "ZccTalker",  # Deprecated on 20230705.
    "ClientConnectorTalker",
    "ZiaTalker",
    "ZpaTalker",
]
